from ..models import User, session_scope, init_database
from ..config import Config
from getpass import getpass
from ..security import encrypt_password

class ResetPasswordCommand():
    def run(self):
        config = Config()
        init_database(config)
        print('Please input username:')
        username = input()

        print('Plase input new password:')
        password = getpass()

        with session_scope() as session:
            user = session.query(User).filter_by(username=username).first()
            if not user:
                print('User does not exist.')
                return

            secret_key = config.get('secret_key')
            encrypted_password = encrypt_password(password, secret_key)
            user.password = encrypted_password
            session.add(user)
            session.commit()

        print('User password updated.')


from sqlalchemy import *
from migrate import *

meta = MetaData()

jobStatus = Table(
    'job_status', meta,
    Column('id', Integer, primary_key=True, autoincrement=False),
    Column('name', String(length=50), nullable=False),
    Column('desc', String(length=200)),
)

def upgrade(migrate_engine):
    meta.bind = migrate_engine
    jobStatus.create()
    migrate_engine.execute(jobStatus.insert().values(id = 0, name='pending', desc='job is not started'))
    migrate_engine.execute(jobStatus.insert().values(id = 1, name='running', desc='job is running'))
    migrate_engine.execute(jobStatus.insert().values(id = 2, name='success', desc='job is finished successfully'))
    migrate_engine.execute(jobStatus.insert().values(id = 3, name='fail', desc='job is finished with error'))
    migrate_engine.execute(jobStatus.insert().values(id = 4, name='warning',desc= 'job is finished with error or warning log'))


def downgrade(migrate_engine):
    meta.bind = migrate_engine
    jobStatus.drop()



from sqlalchemy import *
from migrate import *
from sqlalchemy.ext.declarative import declarative_base

meta = MetaData()
Base = declarative_base(metadata=meta)

class UserKey(Base):
    __tablename__ = 'user_keys'
    id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey('user.id'))
    app_key = Column(String(length=50), nullable=False, unique=True)
    app_secret = Column(String(length=50))
    enabled = Column(Boolean, default=True)
    remark = Column(String(length=255))
    create_at = Column(DateTime)


def upgrade(migrate_engine):
    meta.bind = migrate_engine
    user = Table('user', meta, autoload=True)
    user_key_table = Table('user_keys', meta, autoload=True)
    user_key_table.create()


def downgrade(migrate_engine):
    meta.bind = migrate_engine
    user_key_table = Table('user_keys', meta, autoload=True)
    user_key_table.drop()




from .models import Session, session_scope, SpiderSettings

class SpiderSettingLoader:
    def get_spider_setting(self, spider_id, setting_key):
        with session_scope() as session:
            setting = session.query(SpiderSettings).filter_by(spider_id=spider_id, setting_key=setting_key).first()
            return setting


from pysyncobj import SyncObj, replicated
import logging
from .config import Config

logger = logging.getLogger(__name__)


class ClusterNode():
    def __init__(self, task_id, config=None):
        if config is None:
            config = Config()
        sync_address = config.get('cluster_bind_address')
        sync_ports = config.get('cluster_bind_ports').split(',')
        sync_port = int(sync_ports[task_id])
        peers = config.get('cluster_peers').split(',') if config.get('cluster_peers') else []
        peers = peers + ['%s:%d' % (sync_address, int(port)) for port in sync_ports if int(port) != sync_port]
        logging.debug('starting cluster node.')
        logging.debug('cluster node binding: %s:%d' % (sync_address, sync_port))
        logging.debug('cluster other peers: %s' % peers)
        try:
            self.sync_obj = ClusterSyncObj(sync_address, sync_port, peers)
        except Exception as e:
            logger.error('Error when creating sync_obj')


class ClusterSyncObj(SyncObj):
    def __init__(self, bind_address, bind_port, peers):
        super(ClusterSyncObj, self).__init__('%s:%d' % (bind_address, bind_port), peers)
        self.__counter = 0

    @replicated
    def add_schedule_job(self, trigger_id):
        logger.debug('_add_schedule_job')
        if self.on_add_schedule_job is not None:
            try:
                self.on_add_schedule_job(trigger_id)
            except Exception as e:
                logger.error('Error when adding schedule job : ' + e)

    def set_on_add_schedule_job(self, callback):
        self.on_add_schedule_job = callback

    @replicated
    def remove_schedule_job(self, trigger_id):
        logger.debug('_remove_schedule_job %s' % trigger_id)
        if self.on_remove_schedule_job is not None:
            try:
                self.on_remove_schedule_job(str(trigger_id))
            except Exception as e:
                logger.error('Error when removing schedule job : ' + e)

    def set_on_remove_schedule_job(self, callback):
        self.on_remove_schedule_job = callback



from tornado.ioloop import IOLoop, PeriodicCallback
from tornado.concurrent import Future
from tornado.gen import coroutine
import json
import os
import logging
from .config import AgentConfig
from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPError
from .stream import MultipartRequestBodyProducer
import socket
from tornado import gen
from .workspace import ProjectWorkspace
import tempfile
import shutil
from .exceptions import *
from six.moves.urllib.parse import urlparse, urljoin, urlencode
from six.moves.urllib.error import URLError, HTTPError
from .security import generate_digest
from six.moves.configparser import ConfigParser
from .poster.encode import multipart_encode
from .poster.streaminghttp import register_openers

logger = logging.getLogger(__name__)

TASK_STATUS_SUCCESS = 'success'
TASK_STATUS_FAIL = 'fail'

EXECUTOR_STATUS_OFFLINE = 0
EXECUTOR_STATUS_ONLINE = 1

register_openers()


class NodeAsyncHTTPClient():
    def __init__(self, service_base, io_loop=None, force_instance=False, **kwargs):
        self.service_base = service_base
        self.key = kwargs.pop('key', None)
        self.secret_key = kwargs.pop('secret_key', None)
        self.node_id = None
        self.inner_client = AsyncHTTPClient(io_loop=io_loop, force_instance=force_instance, **kwargs)

    def fetch(self, request, callback=None, raise_error=True, **kwargs):
        if not isinstance(request, HTTPRequest):
            request = HTTPRequest(url=request, **kwargs)
        if self.key and self.secret_key:
            parsed_url = urlparse(request.url)
            path = parsed_url.path
            query = parsed_url.query
            method = request.method
            body = request.body or b''
            request.headers['Authorization'] = '%s %s %s' % ('HMAC',
                                                             self.key,
                                                             generate_digest(self.secret_key, method, path, query,
                                                                             body))
        if self.node_id:
            request.headers['X-Dd-Nodeid'] = str(self.node_id)

        return self.inner_client.fetch(request, callback=callback, raise_error=raise_error, **kwargs)

    @coroutine
    def node_online(self, tags):
        url = urljoin(self.service_base, '/nodes')
        register_postdata = urlencode({'tags': tags})
        request = HTTPRequest(url=url, method='POST', body=register_postdata)

        response = yield self.fetch(request)
        self.node_id = json.loads(response.body)['id']
        raise gen.Return(self.node_id)


class SpiderTask():
    id = None
    spider_id = None
    project_name = None
    spider_name = None
    project_version = None
    spider_parameters = None
    extra_requirements = None


class TaskSlotContainer():
    def __init__(self, max_size=1):
        self._max_size = max_size
        self.slots = [None] * max_size

    def is_full(self):
        for slot in self.slots:
            if slot is None:
                return False
        return True

    def put_task(self, task):
        if not isinstance(task, TaskExecutor):
            raise ValueError('Task in TaskSlotContainer must be TaskExecutor type.')
        for i in range(len(self.slots)):
            if self.slots[i] is None:
                self.slots[i] = task
                break

    def remove_task(self, task):
        for i in range(len(self.slots)):
            if self.slots[i] == task:
                self.slots[i] = None
                break;

    def tasks(self):
        for item in self.slots:
            if item is not None:
                yield item

    def get_task(self, task_id):
        for item in self.slots:
            if item and item.task.id == task_id:
                return item

    @property
    def max_size(self):
        return self._max_size


class Executor():
    heartbeat_interval = 10
    checktask_interval = 10
    custom_ssl_cert = False

    def __init__(self, config=None):
        self.ioloop = IOLoop.current()
        self.node_id = None
        self.status = EXECUTOR_STATUS_OFFLINE
        self.tags = config.get('tags')

        if config is None:
            config = AgentConfig()
        self.task_slots = TaskSlotContainer(config.getint('slots', 1))
        self.config = config
        server_base = config.get('server')
        if urlparse(server_base).scheme == '':
            if config.getint('server_https_port'):
                server_base = 'https://%s:%d' % (server_base, config.getint('server_https_port'))
            else:
                server_base = 'http://%s:%d' % (server_base, config.getint('server_port'))
        self.service_base = server_base
        client_cert = config.get('client_cert') or None
        client_key = config.get('client_key') or None

        httpclient_defaults = {
            'request_timeout': config.getfloat('request_timeout', 60)
        }
        if client_cert:
            httpclient_defaults['client_cert'] = client_cert
        if client_key:
            httpclient_defaults['client_key'] = client_key
        if os.path.exists('keys/ca.crt'):
            self.custom_ssl_cert = True
            httpclient_defaults['ca_certs'] = 'keys/ca.crt'
            httpclient_defaults['validate_cert'] = True
        logger.debug(httpclient_defaults)

        node_key = None
        secret_key = None
        if os.path.exists('conf/node.conf'):
            cp = ConfigParser()
            cp.read('conf/node.conf')
            node_key = cp.get('agent', 'node_key')
            secret_key = cp.get('agent', 'secret_key')

        self.httpclient = NodeAsyncHTTPClient(self.service_base, key=node_key, secret_key=secret_key,
                                              defaults=httpclient_defaults)

    def start(self):
        self.register_node()
        # init heartbeat period callback
        heartbeat_callback = PeriodicCallback(self.send_heartbeat, self.heartbeat_interval * 1000)
        heartbeat_callback.start()

        # init checktask period callback
        # the new version use HEARTBEAT response to tell client whether there is new task on server queue
        # so do not start this period method. But if server is an old version without that header info
        # client sill need to poll GET_TASK.
        self.checktask_callback = PeriodicCallback(self.check_task, self.checktask_interval * 1000)

        self.ioloop.start()

    def check_header_new_task_on_server(self, headers):
        try:
            new_task_on_server = headers['X-DD-New-Task'] == 'True'
            if new_task_on_server:
                self.ioloop.call_later(0, self.check_task)
        except KeyError:
            # if response contains no 'DD-New-Task' header, it might be the old version server,
            # so use the old GET_TASK pool mode
            if not self.checktask_callback.is_running():
                self.checktask_callback.start()

    @coroutine
    def send_heartbeat(self):
        if self.status == EXECUTOR_STATUS_OFFLINE:
            self.register_node()
            return
        url = urljoin(self.service_base, '/nodes/%d/heartbeat' % self.node_id)
        running_tasks = ','.join([task_executor.task.id for task_executor in self.task_slots.tasks()])
        request = HTTPRequest(url=url, method='POST', body='', headers={'X-DD-RunningJobs': running_tasks})
        try:
            res = yield self.httpclient.fetch(request)
            if 'X-DD-KillJobs' in res.headers:
                logger.info('received kill signal %s' % res.headers['X-DD-KillJobs'])
                for job_id in json.loads(res.headers['X-DD-KillJobs']):
                    task_to_kill = self.task_slots.get_task(job_id)
                    if task_to_kill:
                        logger.info('%s' % task_to_kill)
                        task_to_kill.kill()
                        self.task_slots.remove_task(task_to_kill)
            self.check_header_new_task_on_server(res.headers)
        except HTTPError as e:
            if e.code == 400:
                logging.warning('Node expired, register now.')
                self.status = EXECUTOR_STATUS_OFFLINE
                self.register_node()
        except URLError as e:
            logging.warning('Cannot connect to server. %s' % e)
        except HTTPError as e:
            if e.code == 400:
                logging.warning('Node expired, register now.')
                self.status = EXECUTOR_STATUS_OFFLINE
                self.register_node()
        except Exception as e:
            logging.warning('Cannot connect to server. %s' % e)

    @coroutine
    def register_node(self):
        if self.custom_ssl_cert and self.service_base.startswith('https') and not os.path.exists('keys/ca.crt'):
            httpclient = AsyncHTTPClient(force_instance=True)
            cacertrequest = HTTPRequest(urljoin(self.service_base, 'ca.crt'), validate_cert=False)
            try:
                cacertresponse = yield httpclient.fetch(cacertrequest)
                if not os.path.exists('keys'):
                    os.mkdir('keys')
                open('keys/ca.crt', 'wb').write(cacertresponse.body)
            except HTTPError:
                logger.info('Custom ca cert retrieve failed.')
        try:

            node_id = yield self.httpclient.node_online(self.tags)
            self.status = EXECUTOR_STATUS_ONLINE
            self.node_id = node_id
            logger.info('node %d registered' % self.node_id)
        except URLError as e:
            logging.warning('Cannot connect to server, %s' % e)
        except socket.error as e:
            logging.warning('Cannot connect to server, %s' % e)

    def on_new_task_reach(self, task):
        if task is not None:
            task_executor = self.execute_task(task)
            self.task_slots.put_task(task_executor)

    @coroutine
    def get_next_task(self):
        url = urljoin(self.service_base, '/executing/next_task')
        post_data = urlencode({'node_id': self.node_id})
        request = HTTPRequest(url=url, method='POST', body=post_data)
        try:
            response = yield self.httpclient.fetch(request)
            response_content = response.body
            response_data = json.loads(response_content)
            logger.debug(url)
            logger.debug(response_content)
            if response_data['data'] is not None:
                task = SpiderTask()
                task.id = response_data['data']['task']['task_id']
                task.spider_id = response_data['data']['task']['spider_id']
                task.project_name = response_data['data']['task']['project_name']
                task.project_version = response_data['data']['task']['version']
                task.spider_name = response_data['data']['task']['spider_name']
                if 'extra_requirements' in response_data['data']['task'] and \
                        response_data['data']['task']['extra_requirements']:
                    task.extra_requirements = [x for x in
                                               response_data['data']['task']['extra_requirements'].split(';') if x]
                if 'spider_parameters' in response_data['data']['task']:
                    task.spider_parameters = response_data['data']['task']['spider_parameters']
                else:
                    task.spider_parameters = {}
                self.on_new_task_reach(task)
        except URLError:
            logger.warning('Cannot connect to server')

    def execute_task(self, task):
        egg_downloader = ProjectEggDownloader(service_base=self.service_base, client=self.httpclient)
        executor = TaskExecutor(task, egg_downloader=egg_downloader)
        pid = None
        future = executor.execute()
        self.post_start_task(task, pid)
        executor.on_subprocess_start = self.post_start_task
        self.ioloop.add_future(future, self.task_finished)
        return executor

    @coroutine
    def post_start_task(self, task, pid):
        url = urljoin(self.service_base, '/jobs/%s/start' % task.id)
        post_data = urlencode({'pid': pid or ''})
        try:
            request = HTTPRequest(url=url, method='POST', body=post_data)
            respones = yield self.httpclient.fetch(request)
        except URLError as e:
            logger.error('Error when post_task_task: %s' % e)
        except HTTPError as e:
            logger.error('Error when post_task_task: %s' % e)

    def complete_task(self, task_executor, status):
        '''
        @type task_executor: TaskExecutor
        '''
        url = urljoin(self.service_base, '/executing/complete')

        log_file = open(task_executor.output_file, 'rb')

        post_data = {
            'task_id': task_executor.task.id,
            'status': status,
            'log': log_file,
        }

        items_file = None
        if task_executor.items_file and os.path.exists(task_executor.items_file):
            post_data['items'] = items_file = open(task_executor.items_file, "rb")
            logger.debug('item file size : %d' % os.path.getsize(task_executor.items_file))
        logger.debug(post_data)
        datagen, headers = multipart_encode(post_data)
        headers['X-DD-Nodeid'] = str(self.node_id)
        request = HTTPRequest(url, method='POST', headers=headers, body_producer=MultipartRequestBodyProducer(datagen))
        client = self.httpclient
        future = client.fetch(request, raise_error=False)
        self.ioloop.add_future(future, self.complete_task_done(task_executor, log_file, items_file))
        logger.info('task %s finished' % task_executor.task.id)

    def complete_task_done(self, task_executor, log_file, items_file):
        def complete_task_done_f(future):
            '''
            The callback of complete job request, if socket error occurred, retry commiting complete request in 10 seconds.
            If request is completed successfully, remove task from slots.
            Always close stream files.
            @type future: Future
            :return:
            '''
            response = future.result()
            logger.debug(response)
            if log_file:
                log_file.close()
            if items_file:
                items_file.close()
            if response.error and isinstance(response.error, socket.error):
                self.ioloop.call_later(10, self.complete_task, task_executor,
                                       TASK_STATUS_SUCCESS if task_executor.ret_code == 0 else TASK_STATUS_FAIL)
                logger.warning('Socket error when completing job, retry in 10 seconds.')
                return

            if response.error:
                logger.warning('Error when post task complete request: %s' % response.error)
            self.task_slots.remove_task(task_executor)
            logger.debug('complete_task_done')

        return complete_task_done_f

    def task_finished(self, future):
        task_executor = future.result()
        self.complete_task(task_executor, TASK_STATUS_SUCCESS if task_executor.ret_code == 0 else TASK_STATUS_FAIL)

    def check_task(self):
        if not self.task_slots.is_full():
            self.get_next_task()


class TaskExecutor():
    def __init__(self, task, egg_downloader):
        '''
        @type task: SpiderTask
        '''
        self.task = task
        self.egg_downloader = egg_downloader
        self._f_output = None
        self.output_file = None
        self.items_file = None
        self.ret_code = None
        self.workspace_dir = tempfile.mkdtemp(prefix='ddjob-%s-%s-' % (task.project_name, task.id))
        if not os.path.exists(self.workspace_dir):
            os.makedirs(self.workspace_dir)
        self.output_file = str(os.path.join(self.workspace_dir, '%s.log' % self.task.id))
        self._f_output = open(self.output_file, 'w')
        self.on_subprocess_start = None
        self.workspace = ProjectWorkspace(self.task.project_name, base_workdir=self.workspace_dir)

    @gen.coroutine
    def execute(self):
        try:
            yield self.workspace.init()
            logger.info('start fetch spider egg.')
            downloaded_egg = yield self.egg_downloader.download_egg_future(self.task.id)
            with open(downloaded_egg, 'rb') as egg_f:
                self.workspace.put_egg(egg_f, self.task.project_version)
            logger.info('fetch spider egg done.')
            yield self.workspace.install_requirements(self.task.extra_requirements)
            logger.info('start run spider.')
            run_spider_future = self.workspace.run_spider(self.task.spider_name, self.task.spider_parameters,
                                                          f_output=self._f_output, project=self.task.project_name)
            run_spider_pid = self.workspace.processes[0].pid
            if self.on_subprocess_start:
                self.on_subprocess_start(self.task, run_spider_pid)
            self.items_file = yield run_spider_future
            result = self.complete(0)
        except ProcessFailed as e:
            logger.warning('Process Failed when executing task %s: %s' % (self.task.id, e))
            error_log = e.message
            if e.std_output:
                logger.warning(e.std_output)
                error_log += e.std_output
            if e.err_output:
                logger.warning(e.err_output)
                error_log += e.err_output
            result = self.complete_with_error(error_log)
        except Exception as e:
            logger.error('Error when executing task %s: %s' % (self.task.id, e))
            logger.error(e)
            error_log = e.message
            result = self.complete_with_error(error_log)
        raise gen.Return(result)

    def result(self):
        return self

    def complete(self, ret_code):
        self._f_output.close()
        self.ret_code = ret_code
        return self.result()

    def complete_with_error(self, error_message):
        logger.debug(error_message)
        self._f_output.write(error_message)
        self._f_output.close()
        self.ret_code = 1
        return self.result()

    def __del__(self):
        logger.debug('delete task executor for task %s' % self.task.id)
        if self.workspace_dir and os.path.exists(self.workspace_dir):
            shutil.rmtree(self.workspace_dir)

    def kill(self):
        self._f_output.write("Received a kill command, stopping spider.")
        self.workspace.kill_process()


class ProjectEggDownloader(object):
    def __init__(self, service_base, client=None):
        self.download_path = None
        self.service_base = service_base
        if client is None:
            client = AsyncHTTPClient()
        self.client = client

    def download_egg_future(self, job_id):
        ret_future = Future()

        self.download_path = tempfile.mktemp()
        self._fd = open(self.download_path, 'wb')

        logger.debug('begin download egg.')
        egg_request_url = urljoin(self.service_base, '/jobs/%s/egg' % job_id)
        request = HTTPRequest(egg_request_url, streaming_callback=self._handle_chunk)
        client = self.client

        def done_callback(future):
            self._fd.close()
            ex = future.exception()
            if ex:
                return ret_future.set_exception(ex)

            return ret_future.set_result(self.download_path)

        client.fetch(request).add_done_callback(done_callback)
        return ret_future

    def _handle_chunk(self, chunk):
        self._fd.write(chunk)
        self._fd.flush()

    def __del__(self):
        if os.path.exists(self.download_path):
            os.remove(self.download_path)



from sqlalchemy import *
from migrate import *

meta = MetaData()

def upgrade(migrate_engine):
    meta.bind = migrate_engine
    nodes = Table('nodes', meta, autoload=True)
    nodes_tags = Column('tags', String(length=200))
    nodes_tags.create(nodes)

def downgrade(migrate_engine):
    meta.bind = migrate_engine
    nodes = Table('nodes', meta, autoload=True)
    nodes.c['tags'].drop()



from sqlalchemy import *
from migrate import *

meta = MetaData()

spider_parameters = Table(
    'spider_parameters', meta,
    Column('id', Integer, primary_key=True),
    Column('spider_id', Integer),
    Column('parameter_key', String(length=50), nullable=False),
    Column('value', String(length=200)),
)

def upgrade(migrate_engine):
    meta.bind = migrate_engine
    spider_parameters.create()


def downgrade(migrate_engine):
    meta.bind = migrate_engine
    spider_parameters.drop()



from sqlalchemy import *
meta = MetaData()

user = Table(
    'user', meta,
    Column('id', Integer, primary_key=True),
    Column('username', String(length=50), nullable=False, unique=True),
    Column('nickname', String(length=50)),
    Column('password', String(length=50)),
    Column('create_at', DateTime),
    Column('last_login', DateTime),
    Column('is_admin', Boolean, default=False)
)

def upgrade(migrate_engine):
    meta.bind = migrate_engine
    user.create()
    migrate_engine.execute(user.insert().values(id=1, username='admin', password='admin', nickname='admin',
                                                is_admin=True))

def downgrade(migrate_engine):
    meta.bind = migrate_engine
    user.drop()



from .executor import Executor
import logging
import sys
from .daemonize import daemonize
import signal
import os
from .config import AgentConfig
from optparse import OptionParser
import subprocess
import tornado
import scrapydd
from six.moves import input
from six.moves.urllib.parse import urlparse, urljoin, urlencode
from .security import authenticated_request
from tornado.httpclient import HTTPClient, HTTPError
import json
from six.moves.configparser import SafeConfigParser

logger = logging.getLogger(__name__)

def init_logging(config):
    import logging.handlers
    logger = logging.getLogger()
    if not os.path.exists('logs'):
        os.mkdir('logs')
    fh = logging.handlers.TimedRotatingFileHandler('logs/scrapydd-agent.log', when='D', backupCount=7)
    ch = logging.StreamHandler()
    eh = logging.handlers.TimedRotatingFileHandler(os.path.join('logs/scrapydd-agent-error.log'), when='D',
                                                   backupCount=30)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    eh.setFormatter(formatter)
    eh.setLevel(logging.ERROR)

    logger.addHandler(fh)
    logger.addHandler(ch)
    logger.addHandler(eh)

    if config.getboolean('debug'):
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

class Daemon():
    def __init__(self, pidfile):
        self.pidfile = pidfile
        self.subprocess_p = None
        self.pid = 0

    def start_subprocess(self):
        argv = sys.argv
        argv.remove('--daemon')
        pargs = argv
        env = os.environ.copy()
        self.subprocess_p = subprocess.Popen(pargs, env=env)
        self.subprocess_p.wait()

    def read_pidfile(self):
        try:
            with open(self.pidfile, 'r') as f:
                return int(f.readline())
        except IOError:
            return None

    def try_remove_pidfile(self):
        if os.path.exists(self.pidfile):
            os.remove(self.pidfile)

    def on_signal(self, signum, frame):
        print('closing')
        if self.subprocess_p:
            self.subprocess_p.terminate()
        self.try_remove_pidfile()
        tornado.ioloop.IOLoop.instance().stop()

    def start(self):
        print('Starting scrapydd agent daemon.')
        signal.signal(signal.SIGINT, self.on_signal)
        signal.signal(signal.SIGTERM, self.on_signal)
        daemonize(pidfile=self.pidfile)
        start()
        self.try_remove_pidfile()

def run(argv=None):
    if argv is None:
        argv = sys.argv
    parser = OptionParser(prog  = 'scrapydd agent')
    parser.add_option('-g', '--register', action='store_true', help='register this agent.')
    parser.add_option('--daemon', action='store_true', help='run scrapydd agent in daemon mode')
    parser.add_option('--pidfile', help='pid file will be created when daemon started, \
default: scrapydd-agent.pid')
    opts, args = parser.parse_args(argv)

    pidfile = opts.pidfile or 'scrapydd-agent.pid'

    if opts.register:
        run_register()
    elif opts.daemon:
        daemon = Daemon(pidfile=pidfile)
        daemon.start()
    else:
        start()

def run_register():
    key = input('Please input node key:')
    key_secret = input('Please input node key_secret:')
    config = AgentConfig()
    server_base = config.get('server')
    if urlparse(server_base).scheme == '':
        if config.getint('server_https_port'):
            server_base = 'https://%s:%d' % (server_base, config.getint('server_https_port'))
        else:
            server_base = 'http://%s:%d' % (server_base, config.getint('server_port'))

    register_url = urljoin(server_base, '/nodes/register')
    post_data = {'node_key':key}
    request = authenticated_request(url=register_url, method="POST", app_key=key, app_secret=key_secret,
                                    body=urlencode(post_data))
    client = HTTPClient()
    try:
        response = client.fetch(request)
        response_data = json.loads(response.body)
        if not os.path.exists('conf'):
            os.makedirs('conf')
        with open('conf/node.conf', 'w') as f:
            cp = SafeConfigParser()
            cp.add_section('agent')
            cp.set('agent', 'node_id', str(response_data['id']))
            cp.set('agent', 'node_key', key)
            cp.set('agent', 'secret_key', key_secret)
            cp.write(f)
        print('Register succeed!')

    except HTTPError as e:
        print("Error when registering.")
        print(e.message)


def start():
    config = AgentConfig()
    init_logging(config)
    logging.info('------------------------')
    logging.info('Starting scrapydd agent.')
    logging.info('config %s loaded' % config._loaded_files)
    logging.info('scrapydd version : %s' % scrapydd.__version__)
    logging.info('------------------------')
    executor = Executor(config)
    executor.start()

if __name__ == '__main__':
    run()


from .models import Node, Session, session_scope, NodeKey
from tornado.ioloop import IOLoop, PeriodicCallback
import datetime
import logging
from .exceptions import *
import uuid
from .security import generate_random_string

logger = logging.getLogger(__name__)

class NodeManager():
    interval = 10
    node_timeout = 60

    def __init__(self, scheduler_manager):
        self.scheduler_manager = scheduler_manager

    def init(self):
        self.ioloop = IOLoop.current()
        self.peroid_callback = PeriodicCallback(self._poll, self.interval*1000)
        self.peroid_callback.start()

    def _poll(self):
        last_heartbeat_lessthan = datetime.datetime.now() - datetime.timedelta(seconds=self.node_timeout)
        session = Session()
        for node in session.query(Node).filter(Node.last_heartbeat<last_heartbeat_lessthan, Node.isalive==1):
            logger.info('node %d expired' % node.id)
            self.scheduler_manager.on_node_expired(node.id)
            node.isalive = 0
            if not node.node_key_id:
                node.is_deleted = True
            session.add(node)
        session.commit()
        session.close()

        logger.debug('node manager poll')

    def heartbeat(self, node_id):
        try:
            session = Session()
            node = session.query(Node).filter_by(id=node_id).first()
            if node is None:
                raise NodeExpired()
            if node.isalive == 0 and not node.node_key_id:
                raise NodeExpired()
            node.last_heartbeat = datetime.datetime.now()
            node.isalive = True
            session.add(node)
            session.commit()
        finally:
            session.close()

    def create_node(self, remote_ip, tags=None, key_id=None):
        with session_scope() as session:
            node = Node()
            node.client_ip = remote_ip
            node.create_time = datetime.datetime.now()
            node.last_heartbeat = datetime.datetime.now()
            node.isalive = True
            node.tags = tags
            node.node_key_id = key_id
            session.add(node)
            session.commit()
            session.refresh(node)
            return node

    def create_node_key(self):
        with session_scope() as session:
            node_key = NodeKey()
            node_key.key = str(uuid.uuid4())
            node_key.create_at = datetime.datetime.now()
            node_key.secret_key = generate_random_string(32)
            session.add(node_key)
            session.commit()
            session.refresh(node_key)
            return node_key

    def get_node(self, node_id):
        with session_scope() as session:
            return session.query(Node).get(node_id)

    def update_node_use_key(self, node_key, node_id):
        with session_scope() as session:
            node_key.used_node_id = node_id
            session.add(node_key)
            session.commit()



from tornado.ioloop import IOLoop, PeriodicCallback
import json
import urllib
from tornado.httpclient import AsyncHTTPClient, HTTPRequest
import logging
from tornado.concurrent import Future
from .models import Session, WebhookJob, SpiderWebhook, session_scope, SpiderSettings
import os, os.path, shutil
import sys
from .exceptions import *
from six import StringIO

logger = logging.getLogger(__name__)


class WebhookJobStateStorage():
    def start_job(self, job_id):
        raise NotImplementedError()

    def finish_job(self, job_id):
        raise NotImplementedError()

    def get_next_job(self):
        raise NotImplementedError()

    def add_job(self, job_id, payload_url, items_file):
        '''

        @type job_id : str
        @param job_id: The job append to jab queue
        :return:
        '''
        raise NotImplementedError()


class DatabaseWebhookJobStateStorage(WebhookJobStateStorage):
    def __init__(self):
        with session_scope() as session:
            # reset all running jobs to PENDING status on start
            for job in session.query(WebhookJob).filter(WebhookJob.status == 1):
                job.status = 0
                session.add(job)
            session.commit()

    def start_job(self, job_id):
        with session_scope() as session:
            job = session.query(WebhookJob).filter_by(id=job_id).first()
            job.status = 1
            session.add(job)
            session.commit()

    def finish_job(self, job_id, log=None):
        with session_scope() as session:
            job = session.query(WebhookJob).filter_by(id=job_id).first()
            job.status = 2
            if log is not None:
                job.log = log.getvalue()
            session.add(job)
            session.commit()

    def job_failed(self, job_id, log=None):
        with session_scope() as session:
            job = session.query(WebhookJob).filter_by(id=job_id).first()
            job.status = 3
            if log is not None:
                job.log = log.getvalue()
            session.add(job)
            session.commit()

    def get_next_job(self):
        with session_scope() as session:
            next_jobs = list(session.query(WebhookJob).filter(WebhookJob.status == 0).order_by(WebhookJob.id).slice(0, 1))
            if next_jobs:
                return next_jobs[0]

    def add_job(self, job_id, payload_url, items_file, spider_id):
        '''

        @type job_id : str
        @param job_id: The id of the spider job
        @type payload_url : str
        @param payload_url : The target webhook payload url.
        @type items_file : str
        @param items_file : the items jl file path
        :return:
        '''
        with session_scope() as session:
            job = WebhookJob()
            job.job_id = job_id
            job.payload_url = payload_url
            job.items_file = items_file
            job.spider_id = spider_id
            job.status = 0
            session.add(job)



class WebhookJobExecutor():
    def __init__(self, job, item_file, memory_limit, max_batch_size=0):
        self.payload_url = job.payload_url
        self.item_file = item_file
        self.future = Future()
        self.job = job
        self.ioloop = IOLoop.current()
        self.send_msg_interval = 0.1
        self.memory_limit = memory_limit
        self.max_batch_size = max_batch_size
        self.log = StringIO()

    def start(self):
        self.ioloop.call_later(self.send_msg_interval, self.send_next_msg)
        return self.future

    def send_next_msg(self):
        row_count = 0
        rows = []

        max_batch_size = self.max_batch_size
        while True:
            line = self.item_file.readline()
            if line == '':
                break
            try:
                line_data = json.loads(line.decode('utf8'))
                row_count += 1
                rows.append(line_data)
            except ValueError as e:
                logger.error('Error when decoding jl file. %s', e.message)
                self.log.write('Error when decoding jl file. %s\n' % e.message)
                exc = WebhookJobJlDecodeError(executor=self, message='Error when decoding jl file')
                self.on_error(exc)
                return

            if sys.getsizeof(rows) > self.memory_limit:
                message = 'Webhook task memory usage over %s, \n' % self.memory_limit
                self.log.write(message)
                logger.warning(message)
                exc = WebhookJobOverMemoryLimitError(executor=self, message=message)
                self.on_error(exc)
                return

            if max_batch_size > 0 and row_count >= max_batch_size:
                break

        if len(rows) > 0:
            keys = []
            post_dict = {}
            for row in rows:
                # encode unicode fields
                for key, value in row.items():
                    if isinstance(value, unicode):
                        row[key] = value.encode('utf8')

                # fill all keys
                keys = list(set(keys + row.keys()))

            for key in keys:
                post_dict[key] = [row.get(key, '') for row in rows ]

            post_data = urllib.urlencode(post_dict, doseq=True)

            client = AsyncHTTPClient()
            self.log.write('Sending request to %s\n' % self.payload_url)
            request = HTTPRequest(url=self.payload_url,
                                  method='POST',
                                  body=post_data)
            future = client.fetch(request)
            future.add_done_callback(self.schedule_next_send)
        # no data left in file, complete
        else:
            self.finish_job()

    def finish_job(self):
        if self.item_file:
            self.item_file.close()
        self.future.set_result(self.job)


    def on_error(self, exc):
        logging.error('Webhook job failed, %s' % exc)
        self.item_file.close()

        self.future.set_exception(exc)


    def schedule_next_send(self, callback_future=None):
        if callback_future is not None:
            exc = callback_future.exception()
            if exc is not None:
                self.on_error(WebhookJobFailedError(executor=self, message=str(exc), inner_exc=exc))
                return
            response = callback_future.result()
            self.log.write(response.body + '\n')
        self.ioloop.call_later(self.send_msg_interval, self.send_next_msg)


class WebhookDaemon():
    queue_file_dir = 'webhook'

    def __init__(self, config, spider_setting_loader):
        logger.debug('webhookdaemon start')
        self.ioloop = IOLoop.current()
        self.poll_next_job_callback = PeriodicCallback(self.check_job, 10*1000)
        self.current_job = None
        self.storage = DatabaseWebhookJobStateStorage()
        self.poll_next_job_callback.start()
        self.webhook_memory_limit = config.getint('webhook_memory_limit')
        self.spider_setting_loader = spider_setting_loader

        if not os.path.exists(self.queue_file_dir):
            os.makedirs(self.queue_file_dir)

    def check_job(self):
        logger.debug('webhook daemon checkjob.')
        if self.current_job is None:
            next_job = self.storage.get_next_job()
            if next_job:
                items_file = open(next_job.items_file, 'r')

                max_batch_size = 0
                if next_job.spider_id:
                    max_batch_size_setting = self.spider_setting_loader.get_spider_setting(next_job.spider_id, 'webhook_batch_size')
                    if max_batch_size_setting:
                        try:
                            max_batch_size = int(max_batch_size_setting.value)
                        except ValueError:
                            # if setting is not configured correct, set it to 0 by default.
                            max_batch_size = 0

                self.current_job = WebhookJobExecutor(next_job, items_file, self.webhook_memory_limit, max_batch_size=max_batch_size)
                try:
                    future = self.current_job.start()
                    self.storage.start_job(next_job.id)
                    future.add_done_callback(self.job_finised)
                except Exception as e:
                    logger.error('Error when starting webhook job: %s', e.message)
                    self.job_failed(next_job)


    def init(self):
        self.poll_next_job_callback.start()

    def job_finised(self, future):
        try:
            job = future.result()
            log = self.current_job.log
            self.storage.finish_job(job.id, log)
            self.current_job = None
            if os.path.exists(job.items_file) and os.path.dirname(job.items_file) == self.queue_file_dir:
                os.remove(job.items_file)
            logger.info('webhook job %s finished', job.id)
        # except WebhookJobOverMemoryLimitError as e:
        #     job = e.job
        #     self.job_failed(job)
        # except WebhookJobJlDecodeError as e:
        #     job = e.job
        #     self.job_failed(job)
        except WebhookJobFailedError as e:
            executor = e.executor
            job = executor.job
            self.job_failed(job, executor.log)
            logger.error(e)

    def job_failed(self, job, log=None):
        self.storage.job_failed(job.id, log)

        self.current_job = None
        if os.path.exists(job.items_file) and os.path.dirname(job.items_file) == self.queue_file_dir:
            os.remove(job.items_file)
        logger.info('webhook job %s failed', job.id)

    def on_spider_complete(self, job, items_file):
        with session_scope() as session:
            webhook_setting = session.query(SpiderSettings).filter_by(spider_id = job.spider_id, setting_key='webhook_payload').first()
            if webhook_setting and webhook_setting.value:
                webhook_payload_url = webhook_setting.value
                task_items_file = os.path.join(self.queue_file_dir, os.path.basename(items_file))
                shutil.copy(items_file, task_items_file)
                self.storage.add_job(job.id, webhook_payload_url, task_items_file, job.spider_id)



import sys
import os
from scrapy.settings import overridden_settings, Settings, iter_default_settings
from scrapy.utils.project import get_project_settings
from .runner import FilesystemEggStorage, activate_egg, project_environment
import json

def main():
    if len(sys.argv) > 0:
        project = sys.argv[1]
        os.environ['SCRAPY_PROJECT'] = project
    else:
        project = os.environ['SCRAPY_PROJECT']
    with project_environment(project):
        setting = get_project_settings()
        default_settings = dict(iter_default_settings())
        setting_dict = dict(overridden_settings(setting))
        for key in setting.attributes.keys():
            if key not in default_settings:
                setting_dict[key] = setting.attributes[key].value
        print json.dumps(setting_dict)


if __name__ == '__main__':
    main()


import sys
import os
import os.path
from tornado.concurrent import Future
from tornado.ioloop import IOLoop
from tornado import gen
from subprocess import Popen, PIPE
import logging
import tempfile
import shutil
from scrapydd.exceptions import ProcessFailed, InvalidProjectEgg
import json
from zipfile import ZipFile
from w3lib.url import path_to_file_uri
from shutil import copyfileobj
from six import ensure_str

logger = logging.getLogger(__name__)


class ProjectWorkspace(object):
    pip = None
    python = None
    processes = None
    project_workspace_dir = None
    project_check = None
    temp_dir = None

    def __init__(self, project_name, base_workdir=None):
        if not base_workdir:
            base_workdir = tempfile.mkdtemp(prefix='scrapydd-tmp')
        project_workspace_dir = base_workdir
        self.project_workspace_dir = project_workspace_dir
        self.project_name = project_name
        self.processes = []
        if sys.platform.startswith('linux'):
            self.pip = os.path.join(project_workspace_dir, 'bin', 'pip')
            self.python = os.path.join(project_workspace_dir, 'bin', 'python')
        elif sys.platform.startswith('win'):
            self.pip = os.path.join(project_workspace_dir, 'Scripts', 'pip.exe')
            self.python = os.path.join(project_workspace_dir, 'Scripts', 'python.exe')
        else:
            raise NotImplementedError('Unsupported system %s' % sys.platform)

    def init(self):
        """
        init project isolated workspace,
        :return: future
        """
        future = Future()
        if os.path.exists(self.pip) and os.path.exists(self.python):
            future.set_result(self)
            return future

        logger.debug('workspace dir : %s' % (self.project_workspace_dir,))
        logger.info('start creating virtualenv.')
        try:
            process = Popen([sys.executable, '-m', 'virtualenv', '--system-site-packages', self.project_workspace_dir], stdout=PIPE, stderr=PIPE)
            self.processes.append(process)
        except Exception as e:
            future.set_exception(e)
            return future

        def done(process):
            self.processes.remove(process)
            retcode = process.returncode
            if retcode is not None:
                if retcode == 0:
                    logger.info('Create virtualenv done.')
                    future.set_result(self)
                else:
                    std_output = process.stdout.read()
                    err_output = process.stderr.read()
                    future.set_exception(ProcessFailed('Error when init workspace virtualenv ', std_output=std_output, err_output=err_output))

        wait_process(process, done)
        return future

    def find_project_requirements(self, project=None, egg_storage=None, eggf=None):
        project = self.project_name
        if eggf is None:
            eggf = open(os.path.join(self.project_workspace_dir, 'spider.egg'), 'rb')
        with ZipFile(eggf) as egg_zip_file:
            try:
                requires_fileinfo = egg_zip_file.getinfo('EGG-INFO/requires.txt')
                return ensure_str(egg_zip_file.read(requires_fileinfo)).split()
            except KeyError:
                return []

    @gen.coroutine
    def install_requirements(self, extra_requirements=None):
        requirements = self.find_project_requirements(self.project_name)
        requirements += ['scrapydd']
        if extra_requirements:
            requirements += extra_requirements
        logger.info('start install requirements: %s.' % (requirements, ))
        if requirements:
            yield self.pip_install(requirements)
        logger.info('install requirements done.')
        raise gen.Return()

    def pip_install(self, requirements):
        logger.debug('installing requirements: %s' % requirements)
        future = Future()
        try:
            process = Popen([self.pip, 'install'] + requirements, stdout=PIPE, stderr=PIPE)
            self.processes.append(process)
        except Exception as e:
            future.set_exception(e)
            return future

        def done(process):
            self.processes.remove(process)
            retcode = process.returncode
            if retcode is not None:
                if retcode == 0:
                    future.set_result(self)
                else:
                    std_out = process.stdout.read()
                    err_out = process.stderr.read()
                    future.set_exception(ProcessFailed(std_output=std_out, err_output=err_out))

        wait_process(process, done)
        return future

    def spider_list(self):
        future = Future()
        cwd = self.project_workspace_dir
        try:
            env = os.environ.copy()
            env['SCRAPY_PROJECT'] = self.project_name
            env['SCRAPY_EGG'] = 'spider.egg'
            process = Popen([self.python, '-m', 'scrapydd.utils.runner', 'list'], env = env, cwd=cwd, stdout = PIPE, stderr= PIPE)
        except Exception as e:
            logger.error(e)
            future.set_exception(e)
            return future

        def check_process():
            logger.debug('poll')
            retcode = process.poll()
            if retcode is not None:
                if retcode == 0:
                    stdout = process.stdout.read()
                    stdout = ensure_str(stdout)
                    future.set_result(stdout.splitlines())
                else:
                    #future.set_exception(ProcessFailed(std_output=process.stdout.read(), err_output=process.stderr.read()))
                    future.set_exception(InvalidProjectEgg(detail=process.stderr.read()))
                return
            IOLoop.current().call_later(1, check_process)

        check_process()
        return future

    def run_spider(self, spider, spider_parameters=None, f_output=None, project=None):
        ret_future = Future()
        items_file = os.path.join(self.project_workspace_dir, 'items.jl')
        runner = 'scrapydd.utils.runner'
        pargs = [self.python, '-m', runner, 'crawl', spider]
        if project:
            spider_parameters['BOT_NAME'] = project
        if spider_parameters:
            for spider_parameter_key, spider_parameter_value in spider_parameters.items():
                pargs += [
                            '-s',
                            '%s=%s' % (spider_parameter_key, spider_parameter_value)
                          ]
        pargs += ['-o', str(path_to_file_uri(items_file))]

        env = os.environ.copy()
        env['SCRAPY_PROJECT'] = str(self.project_name)
        #env['SCRAPY_JOB'] = str(self.task.id)
        env['SCRAPY_FEED_URI'] = str(path_to_file_uri(items_file))
        env['SCRAPY_EGG'] = 'spider.egg'

        p = Popen(pargs, env = env, stdout = f_output, cwd = self.project_workspace_dir, stderr = f_output)
        self.processes.append(p)

        def done(process):
            self.processes.remove(process)
            if process.returncode:
                return ret_future.set_exception(ProcessFailed())

            return ret_future.set_result(items_file)

        wait_process(p, done)
        return ret_future

    def find_project_settings(self, project):
        cwd = self.project_workspace_dir
        future = Future()
        try:
            env = os.environ.copy()
            env['SCRAPY_PROJECT'] = project
            process = Popen([self.python, '-m', 'scrapydd.utils.extract_settings', project], env = env, cwd=cwd, stdout = PIPE, stderr= PIPE)
            self.processes.append(process)
        except Exception as e:
            logger.error(e)
            future.set_exception(e)
            return future

        def done(process):
            retcode = process.poll()
            self.processes.remove(process)
            if retcode == 0:
                proc_output = process.stdout.read()
                logger.debug('find_project_settings output:')
                logger.debug(proc_output)
                return future.set_result(json.loads(proc_output))
            else:
                # future.set_exception(ProcessFailed(std_output=process.stdout.read(), err_output=process.stderr.read()))
                return future.set_exception(InvalidProjectEgg(detail=process.stderr.read()))

        wait_process(process, done)
        return future

    def put_egg(self, eggfile, version):
        eggfile.seek(0)
        eggpath = os.path.join(self.project_workspace_dir, 'spider.egg')
        with open(eggpath, 'wb') as f:
            copyfileobj(eggfile, f)

    @gen.coroutine
    def kill_process(self, timeout=30):
        logger.info('killing process')
        for process in self.processes:
            if process.poll() is not None:
                continue
            try:
                process.terminate()
            except OSError as e:
                logger.error('Caught OSError when try to terminate subprocess: %s' % e)

            gen.sleep(timeout)

            if process.poll() is not None:
                continue
            try:
                process.kill()
            except OSError as e:
                logger.error('Caught OSError when try to kill subprocess: %s' % e)

    def __del__(self):
        if self.project_workspace_dir and os.path.exists(self.project_workspace_dir):
            shutil.rmtree(self.project_workspace_dir)

def wait_process(process, callback):
    retcode = process.poll()
    if retcode is not None:
        return callback(process)

    IOLoop.current().call_later(1, wait_process, process, callback)


# -*- coding: utf-8 -*-
from six.moves import StringIO
from six import BytesIO
from six import ensure_str
import os
from pkgutil import get_data
from six.moves.configparser import ConfigParser, SafeConfigParser, NoSectionError, NoOptionError
import glob
from os.path import expanduser

def str2bool(v):
  return v.lower() in ("yes", "true", "t", "1", 'on')

class Config(object):
    """A ConfigParser wrapper to support defaults when calling instance
    methods, and also tied to a single section"""

    SECTION = 'server'
    _loaded_files = []

    def __init__(self, values=None, extra_sources=()):
        if values is None:
            sources = self._getsources()
            self.cp = ConfigParser()
            if __package__:
                default_config = ensure_str(get_data(__package__, 'scrapydd.default.conf'))
                self.cp.readfp(StringIO(default_config))
            for source in sources + list(extra_sources):
                if os.path.exists(source):
                    #self._load_config_file(open(source))
                    with open(source, 'r') as f:
                        self.cp.readfp(f)
        else:
            self.cp = SafeConfigParser(values)
            self.cp.add_section(self.SECTION)

    def _getsources(self):
        sources = ['/etc/scrapydd/scrapydd.conf', r'c:\scrapydd\scrapydd.conf']
        sources += sorted(glob.glob('/etc/scrapyd/conf.d/*'))
        sources += ['scrapydd.conf']
        sources += ['conf/scrapydd.conf']
        sources += [expanduser('~/.scrapydd.conf')]
        return sources

    def get(self, option, default=None):
        env_key = 'SCRAPYDD_' + option.replace('.', '_').upper()
        try:
            return os.getenv(env_key) or self.cp.get(self.SECTION, option)
        except (NoSectionError, NoOptionError):
            if default is not None:
                return default
            raise

    def _get(self, option, conv, default=None):
        return conv(self.get(option, default))

    def getint(self, option, default=None):
        return self._get(option, int, default)

    def getboolean(self, option, default=None):
        return self._get(option, str2bool, default)

    def getfloat(self, option, default=None):
        return self._get(option, float, default)


class AgentConfig(Config):
    SECTION = 'agent'



from .base import AppBaseHandler
from tornado.web import authenticated
from ..schedule import JobRunning
import json
from ..models import Project, Spider, session_scope

class RunSpiderHandler(AppBaseHandler):
    def initialize(self, scheduler_manager):
        super(RunSpiderHandler, self).initialize()
        self.scheduler_manager = scheduler_manager

    @authenticated
    def post(self, project_name, spider_name):
        with session_scope() as session:
            project = session.query(Project).filter_by(name=project_name).first()
            if not project:
                return self.set_status(404, 'project not found.')

            spider = session.query(Spider).filter_by(name=spider_name, project_id = project.id).first()
            if not project:
                return self.set_status(404, 'spider not found.')
            try:
                job = self.scheduler_manager.add_task(project_name, spider_name)
                jobid = job.id
                response_data = {
                    'status': 'ok',
                    'jobid': jobid
                }
                self.write(json.dumps(response_data))
            except JobRunning as e:
                response_data = {
                    'status': 'error',
                    'errormsg': 'job is running with jobid %s' % e.jobid
                }
                self.set_status(400, 'job is running')
                self.write(json.dumps(response_data))


from sqlalchemy import *
from migrate import *

meta = MetaData()

def upgrade(migrate_engine):
    meta.bind = migrate_engine
    webhook_jobs = Table('webhook_jobs', meta, autoload=True)
    webhook_jobs_log = Column('log', Text)
    webhook_jobs_log.create(webhook_jobs)


def downgrade(migrate_engine):
    meta.bind = migrate_engine
    webhook_jobs = Table('webhook_jobs', meta, autoload=True)
    webhook_jobs.c['log'].drop()



import sys
import os
import shutil
import tempfile
from contextlib import contextmanager
from ..eggstorage import FilesystemEggStorage
import os
import pkg_resources
from six import next


def activate_egg(eggpath):
    """Activate a Scrapy egg file. This is meant to be used from egg runners
    to activate a Scrapy egg file. Don't use it from other code as it may
    leave unwanted side effects.
    """
    try:
        d = next(pkg_resources.find_distributions(eggpath))
    except StopIteration:
        raise ValueError("Unknown or corrupt egg")
    d.activate()
    settings_module = d.get_entry_info('scrapy', 'settings').module_name
    os.environ.setdefault('SCRAPY_SETTINGS_MODULE', settings_module)


@contextmanager
def project_environment(project):
    eggstorage = FilesystemEggStorage({})
    version, eggfile = eggstorage.get(project)
    if eggfile:
        prefix = '%s-%s-' % (project, version)
        fd, eggpath = tempfile.mkstemp(prefix=prefix, suffix='.egg')
        lf = os.fdopen(fd, 'wb')
        shutil.copyfileobj(eggfile, lf)
        lf.close()
        activate_egg(eggpath)
    else:
        eggpath = None
    try:
        assert 'scrapy.conf' not in sys.modules, "Scrapy settings already loaded"
        yield
    finally:
        if eggpath:
            os.remove(eggpath)


def main():
    from scrapy.cmdline import execute
    from scrapy.settings import Settings
    settings = Settings()
    egg_path = os.environ.get('SCRAPY_EGG')
    if egg_path:
        activate_egg(egg_path)
        settings_module = os.environ.get('SCRAPY_SETTINGS_MODULE')
        if settings_module:
            settings.setmodule(settings_module, priority='project')
        extra_settings_module = os.environ.get('SCRAPY_EXTRA_SETTINGS_MODULE')
        if extra_settings_module:
            settings.setmodule(extra_settings_module)
        return execute(settings=settings)
    project = os.environ['SCRAPY_PROJECT']
    with project_environment(project):
        execute()


if __name__ == '__main__':
    main()



import tornado.web
import tornado.template
import os
from ..models import session_scope, User
from ..security import CookieAuthenticationProvider, HmacAuthorize, BasicAuthentication
import logging
import json

logger = logging.getLogger(__name__)

BASE_DIR = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'templates')


class AppBaseHandler(tornado.web.RequestHandler):
    authentication_providers = [CookieAuthenticationProvider()]

    def get_default_user(self):
        with session_scope() as session:
            return session.query(User).get(1)

    def get_current_user(self):
        if not self.settings.get('enable_authentication', False):
            return self.get_default_user()

        username = None
        for authentication_provider in self.authentication_providers:
            username = authentication_provider.get_user(self)
            if username:
                break

        if not username:
            return None
        with session_scope() as session:
            return session.query(User).filter_by(username=username).first()

    def data_received(self, chunk):
        pass


class RestBaseHandler(AppBaseHandler):
    authentication_providers = [HmacAuthorize(),  BasicAuthentication()]

    def check_xsrf_cookie(self):
        return None

    def send_json(self, data):
        self.set_header('Content-type', 'application/json')
        if isinstance(data, str):
            self.write(data)
        elif isinstance(data, dict):
            self.write(json.dumps(data))


class InputValidationError(Exception):
    def __init__(self, message=None):
        super(InputValidationError, self).__init__()
        self.message = message



from scrapy.utils.python import retry_on_eintr
from scrapy.utils.conf import get_config, closest_scrapy_cfg
import os
import sys
import glob
from subprocess import check_call

_SETUP_PY_TEMPLATE = """
# Automatically created by: scrapydd
from setuptools import setup, find_packages
setup(
    name         = '%(project)s',
    version      = '1.0',
    packages     = find_packages(),
    entry_points = {'scrapy': ['settings = %(settings)s']},
    install_requires = [],
)
""".lstrip()

def _create_default_setup_py(**kwargs):
    with open('setup.py', 'w') as f:
        f.write(_SETUP_PY_TEMPLATE % kwargs)

def _build_egg():
    closest = closest_scrapy_cfg()
    os.chdir(os.path.dirname(closest))
    if not os.path.exists('setup.py'):
        scrapy_project_settings = get_config()
        settings = scrapy_project_settings.get('settings', 'default')
        project = scrapy_project_settings.get('deploy', 'project', 'project')
        _create_default_setup_py(settings=settings, project=project)
    d = 'dist'
    retry_on_eintr(check_call, [sys.executable, 'setup.py', 'clean', '-a', 'bdist_egg', '-d', d],
                   stdout=sys.stdout, stderr=sys.stderr)
    egg = glob.glob(os.path.join(d, '*.egg'))[0]
    return egg, d

class PackageCommand():
    def run(self):
        egg, d = _build_egg()
        print("Egg has been built: %s" % egg)



import tornado.web
import json
from .handlers.base import RestBaseHandler


class RestNodesHandler(RestBaseHandler):
    def initialize(self):
        self.node_manager = self.settings.get('node_manager')

    def post(self):
        tags = self.get_argument('tags', '').strip()
        tags = None if tags == '' else tags
        remote_ip = self.request.remote_ip
        node = self.node_manager.create_node(remote_ip, tags=tags)
        self.write(json.dumps({'id': node.id}))


import os
import pkg_resources
from scrapy.spiderloader import SpiderLoader
from scrapy.settings import Settings
import sys
from six import next


def activate_egg(eggpath):
    """Activate a Scrapy egg file. This is meant to be used from egg runners
    to activate a Scrapy egg file. Don't use it from other code as it may
    leave unwanted side effects.
    """
    try:
        d = next(pkg_resources.find_distributions(eggpath))
    except StopIteration:
        raise ValueError("Unknown or corrupt egg")
    d.activate()
    settings_module = d.get_entry_info('scrapy', 'settings').module_name
    os.environ.setdefault('SCRAPY_SETTINGS_MODULE', settings_module)
    return settings_module


def main(egg_path=None):
    settings_module = activate_egg(egg_path)

    settings = Settings()
    settings.setmodule(settings_module)
    loader = SpiderLoader(settings)
    for spider in loader.list():
        print(spider)


if __name__ == '__main__':
    main(sys.argv[1])



# -*-coding:utf8-*-
import ssl
import tornado.ioloop
import tornado.web
import tornado.template
import tornado.httpserver
import tornado.netutil
from tornado.web import authenticated
from tornado import gen
from .eggstorage import FilesystemEggStorage
from io import StringIO, BytesIO
from .models import Session, Project, Spider, Trigger, SpiderExecutionQueue, Node, init_database, HistoricalJob, \
    SpiderWebhook, session_scope, SpiderSettings, WebhookJob, SpiderParameter, User
from .schedule import SchedulerManager
from .nodes import NodeManager
import datetime
from .config import Config
from .process import fork_processes
import os.path
import sys
import logging
from .exceptions import *
from sqlalchemy import desc
from optparse import OptionParser
import subprocess
import signal
from .stream import PostDataStreamer
from .webhook import WebhookDaemon
from .daemonize import daemonize
from .workspace import ProjectWorkspace
from .cluster import ClusterNode
from .ssl_gen import SSLCertificateGenerator
from .settting import SpiderSettingLoader
from .restapi import *
from .security import NoAuthenticationProvider, CookieAuthenticationProvider, HmacAuthorize
from .handlers.auth import LogoutHandler, SigninHandler, SignupHandler
from .handlers.base import AppBaseHandler
from .handlers.admin import *
from .handlers.profile import *
from .handlers.rest import *
from .handlers.node import NodesHandler, ExecuteNextHandler, ExecuteCompleteHandler, NodeHeartbeatHandler, \
    JobStartHandler, RegisterNodeHandler, JobEggHandler
from .handlers import webui

logger = logging.getLogger(__name__)

BASE_DIR = os.path.dirname(__file__)


class MainHandler(AppBaseHandler):
    @authenticated
    def get(self):
        session = Session()
        projects = list(session.query(Project).order_by(Project.name))
        self.render('index.html', projects=projects)
        session.close()


class UploadProject(AppBaseHandler):
    except_project_settings = ['NEWSPIDER_MODULE',
                               'SPIDER_MODULES',
                               'BOT_NAME',
                               'USER_AGENT',
                               'LOG_LEVEL',
                               'PROJECT',
                               'SETTINGS_MODULE',
                               ]

    @authenticated
    @gen.coroutine
    def post(self):
        project_name = self.get_body_argument('project')
        version = self.get_body_argument('version')
        eggfile = self.request.files['egg'][0]
        eggf = BytesIO(eggfile['body'])

        try:
            workspace = ProjectWorkspace(project_name)

            yield workspace.init()
            workspace.put_egg(eggf, version)
            yield workspace.install_requirements()
            spiders = yield workspace.spider_list()

        except InvalidProjectEgg as e:
            logger.error('Error when uploading project, %s %s' % (e.message, e.detail))
            self.set_status(400, reason=e.message)
            self.finish("<html><title>%(code)d: %(message)s</title>"
                        "<body><pre>%(output)s</pre></body></html>" % {
                            "code": 400,
                            "message": e.message,
                            "output": e.detail,
                        })
            return
        except ProcessFailed as e:
            logger.error('Error when uploading project, %s, %s' % (e.message, e.std_output))
            self.set_status(400, reason=e.message)
            self.finish("<html><title>%(code)d: %(message)s</title>"
                        "<body><pre>%(output)s</pre></body></html>" % {
                            "code": 400,
                            "message": e.message,
                            "output": e.std_output,
                        })
            return

        logger.debug('spiders: %s' % spiders)
        with session_scope() as session:
            storage = FilesystemEggStorage({})
            eggf.seek(0)
            storage.put(eggf, project_name, version)
            project = session.query(Project).filter_by(name=project_name).first()
            if project is None:
                project = Project()
                project.name = project_name
            project.version = version
            session.add(project)
            session.commit()
            session.refresh(project)

            for spider_name in spiders:
                spider = session.query(Spider).filter_by(project_id=project.id, name=spider_name).first()
                if spider is None:
                    spider = Spider()
                    spider.name = spider_name
                    spider.project_id = project.id
                    session.add(spider)
                    session.commit()
                    session.refresh(spider)

                session.commit()

            if self.request.path.endswith('.json'):
                self.write(json.dumps({'status': 'ok', 'spiders': len(spiders)}))
            else:
                self.render("uploadproject.html")

    @authenticated
    def get(self):
        self.render("uploadproject.html")

class AddScheduleHandler(AppBaseHandler):
    def initialize(self, scheduler_manager):
        super(AddScheduleHandler, self).initialize()
        self.scheduler_manager = scheduler_manager

    @authenticated
    def post(self):
        project = self.get_argument('project')
        spider = self.get_argument('spider')
        cron = self.get_argument('cron')

        with session_scope() as session:
            project = session.query(Project).filter(Project.name == project).first()
            spider = session.query(Spider).filter(Spider.project_id == project.id, Spider.name == spider).first()
        try:
            self.scheduler_manager.add_schedule(project, spider, cron)
            response_data = {
                'status': 'ok',
            }
            self.write(json.dumps(response_data))
        except SpiderNotFound:
            response_data = {
                'status': 'error',
                'errormsg': 'spider not found',
            }
            self.write(json.dumps(response_data))
        except ProjectNotFound:
            response_data = {
                'status': 'error',
                'errormsg': 'project not found',
            }
            self.write(json.dumps(response_data))
        except InvalidCronExpression:
            response_data = {
                'status': 'error',
                'errormsg': 'invalid cron expression.',
            }
            self.write(json.dumps(response_data))


class ProjectList(AppBaseHandler):
    @authenticated
    def get(self):
        session = Session()
        projects = session.query(Project)

        response_data = {'projects': {'id': item.id for item in projects}}
        self.write(response_data)
        session.close()


class SpiderInstanceHandler2(AppBaseHandler):
    @authenticated
    def get(self, project, spider):
        session = Session()
        project = session.query(Project).filter(Project.name == project).first()
        spider = session.query(Spider).filter(Spider.project_id == project.id, Spider.name == spider).first()
        jobs = session.query(HistoricalJob) \
            .filter(HistoricalJob.spider_id == spider.id) \
            .order_by(desc(HistoricalJob.start_time)) \
            .slice(0, 100)
        running_jobs = session.query(SpiderExecutionQueue) \
            .filter(SpiderExecutionQueue.spider_id == spider.id) \
            .order_by(desc(SpiderExecutionQueue.update_time))

        jobs_count = session.query(HistoricalJob) \
            .filter(HistoricalJob.spider_id == spider.id).count()


        webhook_jobs = session.query(WebhookJob).filter_by(spider_id=spider.id)

        context = {}
        context['spider'] = spider
        context['project'] = project
        context['jobs'] = jobs
        context['running_jobs'] = running_jobs
        context['settings'] = session.query(SpiderSettings).filter_by(spider_id=spider.id).order_by(
            SpiderSettings.setting_key)
        context['webhook_jobs'] = webhook_jobs
        spider_parameters = session.query(SpiderParameter) \
            .filter_by(spider_id=spider.id) \
            .order_by(SpiderParameter.parameter_key)
        context['spider_parameters'] = {parameter.parameter_key: parameter.value for parameter in spider_parameters}
        self.render("spider.html", jobs_count = jobs_count, **context)
        session.close()


class SpiderEggHandler(AppBaseHandler):
    @authenticated
    def get(self, id):
        session = Session()
        spider = session.query(Spider).filter_by(id=id).first()
        storage = FilesystemEggStorage({})
        version, f = storage.get(spider.project.name)
        self.write(f.read())
        session.close()


class ProjectSpiderEggHandler(AppBaseHandler):
    @authenticated
    def get(self, project, spider):
        with session_scope() as session:
            project = session.query(Project).filter(Project.name == project).first()
            spider = session.query(Spider).filter(Spider.project_id == project.id, Spider.name == spider).first()
            storage = FilesystemEggStorage({})
            version, f = storage.get(spider.project.name)
            self.write(f.read())
            session.close()


class SpiderListHandler(AppBaseHandler):
    @authenticated
    def get(self):
        session = Session()
        spiders = session.query(Spider)
        self.render("spiderlist.html", spiders=spiders)
        session.close()


class SpiderTriggersHandler(AppBaseHandler):
    scheduler_manager = None

    def initialize(self, scheduler_manager=None):
        assert isinstance(scheduler_manager, SchedulerManager)
        assert scheduler_manager
        super(SpiderTriggersHandler, self).initialize()
        self.scheduler_manager = scheduler_manager

    @authenticated
    def get(self, project, spider):
        with session_scope() as session:
            project = session.query(Project).filter(Project.name == project).first()
            spider = session.query(Spider).filter(Spider.project_id == project.id, Spider.name == spider).first()
            context = {'spider': spider, 'errormsg': None}
            self.render("spidercreatetrigger.html", **context)

    @authenticated
    def post(self, project, spider):
        cron = self.get_argument('cron')
        with session_scope() as session:
            project = session.query(Project).filter(Project.name == project).first()

            spider = session.query(Spider).filter(Spider.project_id == project.id, Spider.name == spider).first()
            try:
                self.scheduler_manager.add_schedule(project, spider, cron)
                return self.redirect('/projects/%s/spiders/%s' % (project.name, spider.name))
            except InvalidCronExpression:

                context = {'spider': spider, 'errormsg': 'Invalid cron expression '}
                return self.render("spidercreatetrigger.html", **context)


class DeleteSpiderTriggersHandler(AppBaseHandler):
    scheduler_manager = None

    def initialize(self, scheduler_manager=None):
        super(DeleteSpiderTriggersHandler, self).initialize()
        self.scheduler_manager = scheduler_manager

    @authenticated
    def post(self, project_name, spider_name, trigger_id):
        self.scheduler_manager.remove_schedule(project_name, spider_name, trigger_id)
        self.redirect('/projects/%s/spiders/%s' % (project_name, spider_name))


class DeleteSpiderJobHandler(AppBaseHandler):
    @authenticated
    def post(self, project_name, spider_name, job_id):
        with session_scope() as session:
            project = session.query(Project).filter_by(name=project_name).first()
            if not project:
                return self.set_status(404, 'project not found.')

            spider = session.query(Spider).filter_by(project_id=project.id, name=spider_name).first()
            if not project:
                return self.set_status(404, 'spider not found.')

            job = session.query(HistoricalJob).filter_by(spider_id = spider.id, id=job_id).first()
            if not job:
                return self.set_status(404, 'job not found.')

            if job.log_file:
                try:
                    os.remove(job.log_file)
                except Exception as e:
                    logger.warning(e.message)

            original_log_file = os.path.join('logs', job.project_name, job.spider_name, '%s.log' % job.id)
            if os.path.exists(original_log_file):
                os.remove(original_log_file)

            original_items_file = os.path.join('items', job.project_name, job.spider_name, '%s.jl' % job.id)
            if os.path.exists(original_items_file):
                os.remove(original_items_file)
            session.delete(job)
            session.commit()


class JobsHandler(AppBaseHandler):
    def initialize(self, scheduler_manager):
        super(JobsHandler, self).initialize()
        self.scheduler_manager = scheduler_manager

    @authenticated
    def get(self):
        pending, running, finished = self.scheduler_manager.jobs()
        context = {
            'pending': pending,
            'running': running,
            'finished': finished,
        }
        self.render("jobs.html", **context)


class LogsHandler(AppBaseHandler):
    @authenticated
    def get(self, project, spider, jobid):
        with session_scope() as session:
            job = session.query(HistoricalJob).filter_by(id=jobid).first()
            log_file = job.log_file
            with open(log_file, 'r') as f:
                log = f.read()
            self.render("log.html", log=log)


class ItemsFileHandler(AppBaseHandler):
    @authenticated
    def get(self, project, spider, jobid):
        with session_scope() as session:
            job = session.query(HistoricalJob).filter_by(id=jobid).first()
            items_file = job.items_file
            self.set_header('Content-Type', 'application/json')
            self.write(open(items_file, 'r').read())


class SpiderWebhookHandler(AppBaseHandler):
    @authenticated
    def get(self, project_name, spider_name):
        session = Session()
        project = session.query(Project).filter(Project.name == project_name).first()
        spider = session.query(Spider).filter(Spider.project_id == project.id, Spider.name == spider_name).first()
        webhook_setting = session.query(SpiderSettings).filter_by(spider_id=spider.id,
                                                                  setting_key='webhook_payload').first()
        if webhook_setting:
            self.write(webhook_setting.value)

    @authenticated
    def post(self, project_name, spider_name):
        payload_url = self.get_argument('payload_url')
        with session_scope() as session:
            project = session.query(Project).filter(Project.name == project_name).first()
            spider = session.query(Spider).filter(Spider.project_id == project.id, Spider.name == spider_name).first()
            webhook_setting = session.query(SpiderSettings).filter_by(spider_id=spider.id,
                                                                      setting_key='webhook_payload').first()
            if webhook_setting is None:
                # no existing row
                webhook_setting = SpiderSettings()
                webhook_setting.spider_id = spider.id
                webhook_setting.setting_key = 'webhook_payload'

            webhook_setting.value = payload_url
            session.add(webhook_setting)
            session.commit()

    @authenticated
    def put(self, project_name, spider_name):
        self.post(project_name, spider_name)

    @authenticated
    def delete(self, project_name, spider_name):
        with session_scope() as session:
            project = session.query(Project).filter(Project.name == project_name).first()
            spider = session.query(Spider).filter(Spider.project_id == project.id, Spider.name == spider_name).first()
            session.query(SpiderSettings).filter_by(spider_id=spider.id, setting_key='webhook_payload').delete()
            session.commit()


class DeleteProjectHandler(RestBaseHandler):
    def initialize(self, scheduler_manager):
        super(DeleteProjectHandler, self).initialize()
        self.scheduler_manager = scheduler_manager

    @authenticated
    def post(self):
        project_name = self.get_argument('project')
        with session_scope() as session:
            project = session.query(Project).filter_by(name=project_name).first()
            if not project:
                return self.set_status(404, 'project not found.')
            spiders = session.query(Spider).filter_by(project_id=project.id)
            for spider in spiders:
                triggers = session.query(Trigger).filter_by(spider_id=spider.id)
                session.query(SpiderExecutionQueue).filter_by(spider_id=spider.id).delete()
                session.query(SpiderParameter).filter_by(spider_id=spider.id).delete()
                session.commit()
                for trigger in triggers:
                    self.scheduler_manager.remove_schedule(project_name, spider.name, trigger_id=trigger.id)
                for history_log in session.query(HistoricalJob).filter(HistoricalJob.spider_id == spider.id):
                    try:
                        os.remove(history_log.log_file)
                    except:
                        pass
                    try:
                        os.remove(history_log.items_file)
                    except:
                        pass
                    session.delete(history_log)
                session.delete(spider)
            session.delete(project)
            storage = FilesystemEggStorage({})
            storage.delete(project_name)
        logger.info('project %s deleted' % project_name)


class SpiderSettingsHandler(AppBaseHandler):
    available_settings = {
        'concurrency': '\d+',
        'timeout': '\d+',
        'webhook_payload': '.*',
        'webhook_batch_size': '\d+',
        'tag': '.*',
        'extra_requirements': '.*'
    }

    @authenticated
    def get(self, project, spider):
        with session_scope() as session:
            project = session.query(Project).filter_by(name=project).first()
            spider = session.query(Spider).filter_by(project_id=project.id, name=spider).first()
            job_settings = {setting.setting_key: setting.value for setting in
                            session.query(SpiderSettings).filter_by(spider_id=spider.id)}

            # default setting values
            if 'concurrency' not in job_settings:
                job_settings['concurrency'] = 1
            if 'timeout' not in job_settings:
                job_settings['timeout'] = 3600
            if 'tag' not in job_settings or job_settings['tag'] is None:
                job_settings['tag'] = ''
            context = {}
            context['settings'] = job_settings
            context['project'] = project
            context['spider'] = spider

            spider_parameters = session.query(SpiderParameter) \
                .filter_by(spider_id=spider.id) \
                .order_by(SpiderParameter.parameter_key)
            context['spider_parameters'] = spider_parameters

            return self.render('spidersettings.html', **context)

    @authenticated
    def post(self, project, spider):
        with session_scope() as session:
            project = session.query(Project).filter_by(name=project).first()
            spider = session.query(Spider).filter_by(project_id=project.id, name=spider).first()

            setting_concurrency_value = self.get_body_argument('concurrency', '1')
            setting_concurrency = session.query(SpiderSettings).filter_by(spider_id=spider.id,
                                                                          setting_key='concurrency').first()
            if not setting_concurrency:
                setting_concurrency = SpiderSettings()
                setting_concurrency.spider_id = spider.id
                setting_concurrency.setting_key = 'concurrency'
            setting_concurrency.value = setting_concurrency_value
            session.add(setting_concurrency)

            setting_timeout_value = self.get_body_argument('timeout', '3600')
            setting_timeout = session.query(SpiderSettings).filter_by(spider_id=spider.id,
                                                                      setting_key='timeout').first()
            if not setting_timeout:
                setting_timeout = SpiderSettings()
                setting_timeout.spider_id = spider.id
                setting_timeout.setting_key = 'timeout'
            setting_timeout.value = setting_timeout_value
            session.add(setting_timeout)

            setting_webhook_payload_value = self.get_body_argument('webhook_payload', '')
            setting_webhook_payload = session.query(SpiderSettings).filter_by(spider_id=spider.id,
                                                                              setting_key='webhook_payload').first()
            if not setting_webhook_payload:
                setting_webhook_payload = SpiderSettings()
                setting_webhook_payload.spider_id = spider.id
                setting_webhook_payload.setting_key = 'webhook_payload'
            setting_webhook_payload.value = setting_webhook_payload_value
            session.add(setting_webhook_payload)

            setting_webhook_batch_size_value = self.get_body_argument('webhook_batch_size', '')
            setting_webhook_batch_size = session.query(SpiderSettings).filter_by(spider_id=spider.id,
                                                                                 setting_key='webhook_batch_size').first()
            if not setting_webhook_batch_size:
                setting_webhook_batch_size = SpiderSettings()
                setting_webhook_batch_size.spider_id = spider.id
                setting_webhook_batch_size.setting_key = 'webhook_batch_size'
            setting_webhook_batch_size.value = setting_webhook_batch_size_value
            session.add(setting_webhook_batch_size)

            setting_tag_value = self.get_body_argument('tag', '').strip()
            setting_tag_value = None if setting_tag_value == '' else setting_tag_value
            setting_tag = session.query(SpiderSettings).filter_by(spider_id=spider.id,
                                                                  setting_key='tag').first()
            if not setting_tag:
                setting_tag = SpiderSettings()
                setting_tag.spider_id = spider.id
                setting_tag.setting_key = 'tag'
            setting_tag.value = setting_tag_value
            session.add(setting_tag)

            setting_extra_requirements_value = self.get_body_argument('extra_requirements', '').strip()
            setting_extra_requirements_value = None if setting_extra_requirements_value == '' else setting_extra_requirements_value
            setting_extra_requirements = session.query(SpiderSettings).filter_by(spider_id=spider.id,
                                                                                 setting_key='extra_requirements').first()
            if not setting_extra_requirements:
                setting_extra_requirements = SpiderSettings()
                setting_extra_requirements.spider_id = spider.id
                setting_extra_requirements.setting_key = 'extra_requirements'
            setting_extra_requirements.value = setting_extra_requirements_value
            session.add(setting_extra_requirements)

            spider_parameter_keys = self.get_body_arguments('SpiderParameterKey')
            spider_parameter_values = self.get_body_arguments('SpiderParameterValue')
            session.query(SpiderParameter).filter_by(spider_id=spider.id).delete()
            for i, spider_parameter_key in enumerate(spider_parameter_keys):
                spider_parameter_key = spider_parameter_key.strip()
                if spider_parameter_key == '':
                    continue

                spider_parameter_value = spider_parameter_values[i]
                spider_parameter = session.query(SpiderParameter).filter_by(spider_id=spider.id,
                                                                            parameter_key=spider_parameter_key).first()
                if not spider_parameter:
                    spider_parameter = SpiderParameter()
                    spider_parameter.spider_id = spider.id
                    spider_parameter.parameter_key = spider_parameter_key
                spider_parameter.value = spider_parameter_value
                session.add(spider_parameter)

            self.redirect('/projects/%s/spiders/%s' % (project.name, spider.name))


class CACertHandler(RestBaseHandler):
    def get(self):
        self.write(open('keys/ca.crt', 'rb').read())
        self.set_header('Content-Type', 'application/cert')


class ListProjectVersionsHandler(RestBaseHandler):
    @authenticated
    def get(self):
        try:
            project = self.get_argument('project')
        except tornado.web.MissingArgumentError as e:
            return self.write({'status': 'error', 'message': e.arg_name})

        storage = FilesystemEggStorage({})
        versions = storage.list(project)
        return self.write({'status': 'ok', 'versions': versions})


def make_app(scheduler_manager, node_manager, webhook_daemon=None, authentication_providers=None, debug=False,
             enable_authentication=False, secret_key='', enable_node_registration=False):
    """

    @type scheduler_manager SchedulerManager
    @type node_manager NodeManager
    @type webhook_daemon: WebhookDaemon
    @type authentication_provider: AuthenticationProvider

    :return:
    """

    settings = dict(cookie_secret=secret_key,
                    login_url="/signin",
                    static_path=os.path.join(BASE_DIR, 'static'),
                    template_path=os.path.join(BASE_DIR, 'templates'),
                    xsrf_cookies=True,
                    debug=debug,
                    enable_authentication=enable_authentication,
                    scheduler_manager=scheduler_manager,
                    enable_node_registration=enable_node_registration)

    if authentication_providers is None:
        authentication_providers = []

    if len(authentication_providers) == 0:
        authentication_providers.append(NoAuthenticationProvider())

    settings['authentication_providers'] = authentication_providers
    settings['node_manager'] = node_manager

    return tornado.web.Application([
        (r"/", MainHandler),
        (r'/signin', SigninHandler),
        (r'/logout', LogoutHandler),
        (r'/uploadproject', UploadProject),

        # scrapyd apis
        (r'/addversion.json', AddVersionHandler),
        (r'/delproject.json', DeleteProjectHandler, {'scheduler_manager': scheduler_manager}),
        (r'/listversions.json', ListProjectVersionsHandler),
        (r'/schedule.json', ScheduleHandler, {'scheduler_manager': scheduler_manager}),

        (r'/add_schedule.json', AddScheduleHandler, {'scheduler_manager': scheduler_manager}),
        (r'/projects', ProjectList),
        (r'/spiders', SpiderListHandler),
        (r'/spiders/(\d+)/egg', SpiderEggHandler),
        (r'/projects/(\w+)/spiders/(\w+)', SpiderInstanceHandler2),
        (r'/projects/(\w+)/spiders/(\w+)/triggers', SpiderTriggersHandler, {'scheduler_manager': scheduler_manager}),
        (r'/projects/(\w+)/spiders/(\w+)/triggers/(\w+)/delete', DeleteSpiderTriggersHandler,
            {'scheduler_manager': scheduler_manager}),
        (r'/projects/(\w+)/spiders/(\w+)/jobs/(\w+)/delete', DeleteSpiderJobHandler),
        (r'/projects/(\w+)/spiders/(\w+)/settings', SpiderSettingsHandler),
        (r'/projects/(\w+)/spiders/(\w+)/webhook', SpiderWebhookHandler),
        (r'^/projects/(\w+)/spiders/(\w+)/run$', webui.RunSpiderHandler, {'scheduler_manager': scheduler_manager}),
        (r'/projects/(\w+)/spiders/(\w+)/egg', ProjectSpiderEggHandler),

        (r'/profile$', ProfileHomeHandler),
        (r'/profile/keys$', ProfileKeysHandler),
        (r'/profile/change_password$', ProfileChangepasswordHandler),

        (r'/admin$', AdminHomeHandler),
        (r'^/admin/nodes$', AdminNodesHandler),

        # agent node ysing handlers
        (r'/executing/next_task', ExecuteNextHandler, {'scheduler_manager': scheduler_manager}),
        (r'/executing/complete', ExecuteCompleteHandler,
         {'webhook_daemon': webhook_daemon, 'scheduler_manager': scheduler_manager}),
        (r'/nodes', NodesHandler, {'node_manager': node_manager}),
        (r'/nodes/register', RegisterNodeHandler),

        (r'/nodes/(\d+)/heartbeat', NodeHeartbeatHandler,
         {'node_manager': node_manager, 'scheduler_manager': scheduler_manager}),
        (r'/jobs', JobsHandler, {'scheduler_manager': scheduler_manager}),
        (r'/jobs/(\w+)/start', JobStartHandler, {'scheduler_manager': scheduler_manager}),
        (r'/jobs/(\w+)/egg', JobEggHandler),
        (r'/logs/(\w+)/(\w+)/(\w+).log', LogsHandler),
        (r'/items/(\w+)/(\w+)/(\w+).jl', ItemsFileHandler),
        (r'/ca.crt', CACertHandler),
        (r'/static/(.*)', tornado.web.StaticFileHandler, {'path': os.path.join(os.path.dirname(__file__), 'static')}),

        (r'/api/v1/nodes$', RestNodesHandler),
        (r'/api/v1/jobs/next', GetNextJobHandler, {'scheduler_manager': scheduler_manager}),
        (r'/api/v1/nodes/register', RestRegisterNodeHandler),
    ], **settings)


def check_and_gen_ssl_keys(config):
    server_name = config.get('server_name')
    if not server_name:
        raise Exception('Must specify a server name')
    ssl_gen = SSLCertificateGenerator()
    try:
        ca_key = ssl_gen.get_ca_key()
        ca_cert = ssl_gen.get_ca_cert()
    except IOError:
        logger.info('ca cert not exist, creating new cert and key.')
        ssl_gen.gen_ca('scrapydd', 'scrapydd')

    host_name = config.get('server_name')
    if not os.path.exists(os.path.join('keys', '%s.crt' % host_name)):
        logger.info('server cert not exist, creating new.')
        alt_names = [s.strip() for s in config.get('dns_alt_names').split(',')]
        if '' in alt_names:
            alt_names.remove('')
        ssl_gen.gen_cert(host_name, usage=2, alt_names=alt_names)


def start_server(argv=None):
    config = Config()
    logging.debug('starting server with argv : %s' % str(argv))

    is_debug=config.getboolean('debug')
    init_database()
    bind_address = config.get('bind_address')
    bind_port = config.getint('bind_port')
    try:
        https_port = int(config.get('https_port')) if config.get('https_port') else None
    except ValueError:
        logger.warning('https_port is configured, but it is not int, %s' % config.get('https_port'))
        https_port = None

    if https_port:
        https_sockets = tornado.netutil.bind_sockets(https_port, bind_address)

    sockets = tornado.netutil.bind_sockets(bind_port, bind_address)

    # task_id is current process identifier when forked processes, start with 0
    task_id = None
    if not is_debug:
        if not sys.platform.startswith('win'):
            task_id = fork_processes(config.getint('fork_proc_count'))
        else:
            logger.warning('Windows platform does not support forking process, running in single process mode.')

    cluster_sync_obj = None
    if task_id is not None and config.get('cluster_bind_address'):
        cluster_node = ClusterNode(task_id, config)
        cluster_sync_obj = cluster_node.sync_obj

    scheduler_manager = SchedulerManager(config=config, syncobj=cluster_sync_obj)
    scheduler_manager.init()

    node_manager = NodeManager(scheduler_manager)
    node_manager.init()

    webhook_daemon = WebhookDaemon(config, SpiderSettingLoader())
    webhook_daemon.init()
    enable_authentication = config.getboolean('enable_authentication')
    secret_key = config.get('secret_key')
    app = make_app(scheduler_manager, node_manager, webhook_daemon,
                   debug=is_debug,
                   enable_authentication=enable_authentication,
                   secret_key=secret_key,
                   enable_node_registration=config.getboolean('enable_node_registration', False))

    server = tornado.httpserver.HTTPServer(app)
    server.add_sockets(sockets)

    if https_port:
        check_and_gen_ssl_keys(config)
        if config.getboolean('client_validation'):
            ssl_ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        else:
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_ctx.load_cert_chain(os.path.join('keys', "%s.crt" % config.get('server_name')),
                                os.path.join('keys', "%s.key" % config.get('server_name')))
        ssl_ctx.load_verify_locations(cafile='keys/ca.crt')
        ssl_ctx.check_hostname = False
        httpsserver = tornado.httpserver.HTTPServer(app, ssl_options=ssl_ctx)
        httpsserver.add_sockets(https_sockets)
        logger.info('starting https server on %s:%s' % (bind_address, https_port))
    ioloop = tornado.ioloop.IOLoop.current()
    ioloop.start()


def run(argv=None):
    if argv is None:
        argv = sys.argv
    parser = OptionParser(prog='scrapydd server')
    parser.add_option('--daemon', action='store_true', help='run scrapydd server in daemon mode')
    parser.add_option('--pidfile', help='pid file will be created when daemon started')
    opts, args = parser.parse_args(argv)
    pidfile = opts.pidfile or 'scrapydd-server.pid'

    config = Config()
    init_logging(config)

    if opts.daemon:
        print('starting daemon.')
        daemon = Daemon(pidfile=pidfile)
        daemon.start()
        sys.exit(0)
    else:
        start_server()


def init_logging(config):
    import logging.handlers
    log_dir = 'logs'
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    logger = logging.getLogger()
    fh = logging.handlers.TimedRotatingFileHandler(os.path.join(log_dir, 'scrapydd-server.log'), when='D',
                                                   backupCount=7)
    eh = logging.handlers.TimedRotatingFileHandler(os.path.join(log_dir, 'scrapydd-error.log'), when='D',
                                                   backupCount=30)
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    eh.setFormatter(formatter)
    eh.setLevel(logging.ERROR)
    logger.addHandler(fh)
    logger.addHandler(ch)
    logger.addHandler(eh)

    if config.getboolean('debug'):
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    access_log_logger = logging.getLogger('tornado.access')
    access_log_fh = logging.handlers.TimedRotatingFileHandler(os.path.join(log_dir, 'scrapydd-access.log'), when='D',
                                                              backupCount=30)
    access_log_logger.addHandler(access_log_fh)
    access_log_logger.setLevel(logging.INFO)


class Daemon():
    def __init__(self, pidfile):
        self.pidfile = pidfile
        self.subprocess_p = None
        self.pid = 0

    def start_subprocess(self):
        argv = sys.argv
        argv.remove('--daemon')
        pargs = argv
        env = os.environ.copy()
        self.subprocess_p = subprocess.Popen(pargs, env=env)
        signal.signal(signal.SIGINT, self.on_signal)
        signal.signal(signal.SIGTERM, self.on_signal)
        self.subprocess_p.wait()

    def read_pidfile(self):
        try:
            with open(self.pidfile, 'r') as f:
                return int(f.readline())
        except IOError:
            return None

    def try_remove_pidfile(self):
        if os.path.exists(self.pidfile):
            os.remove(self.pidfile)

    def on_signal(self, signum, frame):
        logger.info('receive signal %d closing' % signum)
        if self.subprocess_p:
            self.subprocess_p.terminate()
        self.try_remove_pidfile()
        tornado.ioloop.IOLoop.instance().stop()

    def start(self):
        # signal.signal(signal.SIGINT, self.on_signal)
        # signal.signal(signal.SIGTERM, self.on_signal)
        daemonize(pidfile=self.pidfile)
        # self.start_subprocess()
        start_server()
        self.try_remove_pidfile()


if __name__ == "__main__":
    run()



import re
from glob import glob
from os import path, makedirs, remove
from shutil import copyfileobj, rmtree
from distutils.version import LooseVersion

class FilesystemEggStorage(object):

    def __init__(self, config):
        self.basedir = config.get('eggs_dir', 'eggs')

    def put(self, eggfile, project, version):
        eggpath = self._eggpath(project, version)
        eggdir = path.dirname(eggpath)
        if not path.exists(eggdir):
            makedirs(eggdir)
        with open(eggpath, 'wb') as f:
            copyfileobj(eggfile, f)

    def get(self, project, version=None):
        if version is None:
            try:
                version = self.list(project)[-1]
            except IndexError:
                return None, None
        return version, open(self._eggpath(project, version), 'rb')

    def list(self, project):
        eggdir = path.join(self.basedir, project)
        versions = [path.splitext(path.basename(x))[0] \
            for x in glob("%s/*.egg" % eggdir)]
        return sorted(versions, key=LooseVersion)

    def delete(self, project, version=None):
        if version is None:
            rmtree(path.join(self.basedir, project))
        else:
            remove(self._eggpath(project, version))
            if not self.list(project): # remove project if no versions left
                self.delete(project)

    def _eggpath(self, project, version):
        sanitized_version = re.sub(r'[^a-zA-Z0-9_-]', '_', version)
        x = path.join(self.basedir, project, "%s.egg" % sanitized_version)
        return x



from .base import RestBaseHandler
from ..models import session_scope, Session, Spider, Project, SpiderExecutionQueue, SpiderSettings, NodeKey
import json
import logging
import tornado
import datetime
import os
from ..stream import PostDataStreamer
from ..exceptions import *
from six import ensure_str, binary_type
from six.moves.urllib.parse import urlparse, urlencode
from ..security import generate_digest
from tornado.web import HTTPError
import hmac
from ..eggstorage import FilesystemEggStorage
import functools

logger = logging.getLogger(__name__)


def authenticated(method):
    """Decorate methods with this to require that the user be logged in.

    If the user is not logged in, they will be redirected to the configured
    `login url <RequestHandler.get_login_url>`.

    If you configure a login url with a query parameter, Tornado will
    assume you know what you're doing and use it as-is.  If not, it
    will add a `next` parameter so the login page knows where to send
    you once you're logged in.
    """
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self.current_user:
            raise HTTPError(403)
        return method(self, *args, **kwargs)
    return wrapper

class NodeHmacAuthenticationProvider(object):
    def get_user(self, handler):
        authorization = handler.request.headers.get("Authorization", "").split(" ")
        if len(authorization) != 3:
            logging.info("Invalid Authorization header {}".format(authorization))
            return None

        algorithm, key, provided_digest = authorization
        if algorithm != "HMAC":
            logging.info("Invalid algorithm {}".format(algorithm))
            return None

        with session_scope() as session:
            node_key = session.query(NodeKey).filter_by(key=key).first()

            if node_key is None:
                logging.info("Invalid HMAC key {}".format(key))
                return None
            secret = node_key.secret_key
            body = handler.request.body if isinstance(handler.request.body, binary_type) else b''
            expected_digest = generate_digest(
                secret, handler.request.method, handler.request.path, handler.request.query,
                body)


            if not hmac.compare_digest(expected_digest, provided_digest):
                logging.info("Invalid HMAC digest {}".format(provided_digest))
                return None

            return node_key.used_node_id



class NodeBaseHandler(RestBaseHandler):
    authentication_providers = [NodeHmacAuthenticationProvider()]

    def get_current_user(self):
        # if enable_node_registration is not enabled, node do not need register to work.
        # get node_id from request
        if not self.settings.get('enable_node_registration', False):
            return int(self.request.headers.get('X-Dd-Nodeid'))

        for authentication_provider in self.authentication_providers:
            node_id = authentication_provider.get_user(self)
            return node_id


class RegisterNodeHandler(NodeBaseHandler):
    @authenticated
    def post(self):
        node_key = self.current_user
        with session_scope() as session:
            tags = self.get_argument('tags', '').strip()
            tags = None if tags == '' else tags
            remote_ip = self.request.remote_ip
            node_manager = self.settings.get('node_manager')
            node = node_manager.create_node(remote_ip, tags=tags, key_id=node_key.id)
            node_key.used_node_id = node.id
            session.add(node_key)
            session.commit()
            self.send_json({'id': node.id})

    def get_current_user(self):
        authorization = self.request.headers.get("Authorization", "").split(" ")
        if len(authorization) != 3:
            logging.info("Invalid Authorization header {}".format(authorization))
            return None

        algorithm, key, provided_digest = authorization
        if algorithm != "HMAC":
            logging.info("Invalid algorithm {}".format(algorithm))
            return None

        with session_scope() as session:
            user_key = session.query(NodeKey).filter_by(key=key).first()

            if user_key is None:
                logging.info("Invalid HMAC key {}".format(key))
                return None

            if user_key.is_deleted:
                logging.info("Invalid HMAC key {}".format(key))
                return None

            secret = user_key.secret_key
            expected_digest = generate_digest(
                secret, self.request.method, self.request.path, self.request.query,
                self.request.body)

            if not hmac.compare_digest(expected_digest, provided_digest):
                logging.info("Invalid HMAC digest {}".format(provided_digest))
                return None

            return user_key


class NodesHandler(NodeBaseHandler):
    def initialize(self, node_manager):
        super(NodesHandler, self).initialize()
        self.node_manager = node_manager

    def post(self):
        enable_node_registration = self.settings.get('enable_node_registration', False)
        node_id = NodeHmacAuthenticationProvider().get_user(self)

        # if node_registerion is enabled, node should be assigned node_id before use by
        # invoke /nodes/register
        logger.debug('enable_node_registration : %s' % enable_node_registration)
        if enable_node_registration and node_id is None:
            return self.set_status(403, 'node registration is enabled, no key provided.')

        tags = self.get_argument('tags', '').strip()
        tags = None if tags == '' else tags
        remote_ip = self.request.headers.get('X-Real-IP') or self.request.remote_ip
        if node_id:
            node = self.node_manager.get_node(node_id)
            with session_scope() as session:
                node.tags = tags
                node.isalive = True
                node.last_heartbeat = datetime.datetime.now()
                session.add(node)
                session.commit()
        else:
            node = self.node_manager.create_node(remote_ip, tags=tags)
        self.write(json.dumps({'id': node.id}))


class ExecuteNextHandler(NodeBaseHandler):
    scheduler_manager = None

    def initialize(self, scheduler_manager=None):
        self.scheduler_manager = scheduler_manager

    @authenticated
    def post(self):
        with session_scope() as session:
            node_id = int(self.get_argument('node_id'))
            next_task = self.scheduler_manager.get_next_task(node_id)

            response_data = {'data': None}

            if next_task is not None:
                spider = session.query(Spider).filter_by(id=next_task.spider_id).first()
                if not spider:
                    logger.error('Task %s has not spider, deleting.' % next_task.id)
                    session.query(SpiderExecutionQueue).filter_by(id=next_task.id).delete()
                    self.write({'data': None})
                    return

                project = session.query(Project).filter_by(id=spider.project_id).first()
                if not project:
                    logger.error('Task %s has not project, deleting.' % next_task.id)
                    session.query(SpiderExecutionQueue).filter_by(id=next_task.id).delete()
                    self.write({'data': None})
                    return

                extra_requirements_setting = session.query(SpiderSettings) \
                    .filter_by(spider_id=spider.id, setting_key='extra_requirements').first()

                extra_requirements = extra_requirements_setting.value if extra_requirements_setting else ''
                response_data['data'] = {'task': {
                    'task_id': next_task.id,
                    'spider_id': next_task.spider_id,
                    'spider_name': next_task.spider_name,
                    'project_name': next_task.project_name,
                    'version': project.version,
                    'extra_requirements': extra_requirements,
                    'spider_parameters': {parameter.parameter_key: parameter.value for parameter in spider.parameters}
                }}
            self.write(json.dumps(response_data))


MB = 1024 * 1024
GB = 1024 * MB
TB = 1024 * GB
MAX_STREAMED_SIZE = 1 * GB


@tornado.web.stream_request_body
class ExecuteCompleteHandler(NodeBaseHandler):
    webhook_daemon = None
    scheduler_manager = None
    ps = None

    def initialize(self, webhook_daemon=None, scheduler_manager=None):
        """

        @type webhook_daemon : WebhookDaemon

        @type scheduler_manager: SchedulerManager
        :return:
        """
        super(ExecuteCompleteHandler, self).initialize()
        self.webhook_daemon = webhook_daemon
        self.scheduler_manager = scheduler_manager

    def prepare(self):
        # set the max size limiation here
        self.request.connection.set_max_body_size(MAX_STREAMED_SIZE)
        try:
            total = int(self.request.headers.get("Content-Length", "0"))
        except TypeError:
            total = 0
        self.ps = PostDataStreamer(total)  # ,tmpdir="/tmp"

        # self.fout = open("raw_received.dat","wb+")

    @authenticated
    def post(self):
        try:
            self.ps.finish_receive()
            fields = self.ps.get_values(['task_id', 'status'])
            logger.debug(self.ps.get_nonfile_names())
            node_id = self.request.headers.get('X-Dd-Nodeid')
            task_id = ensure_str(fields['task_id'])
            status = ensure_str(fields['status'])
            if status == 'success':
                status_int = 2
            elif status == 'fail':
                status_int = 3
            else:
                self.set_status(401, 'Invalid argument: status.')
                return

            session = Session()
            query = session.query(SpiderExecutionQueue) \
                .filter(SpiderExecutionQueue.id == task_id, SpiderExecutionQueue.status.in_([1,5]))
            # be compatible with old agent version
            if node_id:
                query = query.filter(SpiderExecutionQueue.node_id == node_id)
            else:
                logger.warning('Agent has not specified node id in complete request, client address: %s.' %
                               self.request.remote_ip)
            job = query.first()

            if job is None:
                self.set_status(404, 'Job not found.')
                session.close()
                return
            log_file = None
            items_file = None
            try:
                spider_log_folder = os.path.join('logs', job.project_name, job.spider_name)
                if not os.path.exists(spider_log_folder):
                    os.makedirs(spider_log_folder)

                log_part = self.ps.get_parts_by_name('log')[0]
                if log_part:
                    import shutil
                    log_file = os.path.join(spider_log_folder, job.id + '.log')
                    shutil.copy(log_part['tmpfile'].name, log_file)

            except Exception as e:
                logger.error('Error when writing task log file, %s' % e)

            items_parts = self.ps.get_parts_by_name('items')
            if items_parts:
                try:
                    part = items_parts[0]
                    tmpfile = part['tmpfile'].name
                    logger.debug('tmpfile size: %d' % os.path.getsize(tmpfile))
                    items_file_path = os.path.join('items', job.project_name, job.spider_name)
                    if not os.path.exists(items_file_path):
                        os.makedirs(items_file_path)
                    items_file = os.path.join(items_file_path, '%s.jl' % job.id)
                    import shutil
                    shutil.copy(tmpfile, items_file)
                    logger.debug('item file size: %d' % os.path.getsize(items_file))
                except Exception as e:
                    logger.error('Error when writing items file, %s' % e)

            job.status = status_int
            job.update_time = datetime.datetime.now()
            historical_job = self.scheduler_manager.job_finished(job, log_file, items_file)

            if items_file:
                self.webhook_daemon.on_spider_complete(historical_job, items_file)

            session.close()
            logger.info('Job %s completed.' % task_id)
            response_data = {'status': 'ok'}
            self.write(json.dumps(response_data))

        finally:
            # Don't forget to release temporary files.
            self.ps.release_parts()

    def data_received(self, chunk):
        self.ps.receive(chunk)


class NodeHeartbeatHandler(NodeBaseHandler):
    def initialize(self, node_manager, scheduler_manager):
        super(NodeHeartbeatHandler, self).initialize()
        self.node_manager = node_manager
        self.scheduler_manager = scheduler_manager

    @authenticated
    def post(self, id):
        # logger.debug(self.request.headers)
        node_id = int(self.current_user)
        self.set_header('X-DD-New-Task', self.scheduler_manager.has_task(node_id))
        try:
            self.node_manager.heartbeat(node_id)
            running_jobs = self.request.headers.get('X-DD-RunningJobs', None)
            if running_jobs:
                killing_jobs = list(self.scheduler_manager.jobs_running(node_id, running_jobs.split(',')))
                if killing_jobs:
                    logger.info('killing %s' % killing_jobs)
                    self.set_header('X-DD-KillJobs', json.dumps(killing_jobs))
            response_data = {'status': 'ok'}
        except NodeExpired:
            response_data = {'status': 'error', 'errmsg': 'Node expired'}
            self.set_status(400, 'Node expired')
        self.write(json.dumps(response_data))


class JobStartHandler(NodeBaseHandler):
    def initialize(self, scheduler_manager):
        super(JobStartHandler, self).initialize()
        self.scheduler_manager = scheduler_manager

    @authenticated
    def post(self, jobid):
        pid = self.get_argument('pid')
        pid = int(pid) if pid else None
        self.scheduler_manager.job_start(jobid, pid)


class JobEggHandler(NodeBaseHandler):
    @authenticated
    def get(self, jobid):
        with session_scope() as session:
            job = session.query(SpiderExecutionQueue).filter_by(id=jobid).first()
            if not job:
                raise tornado.web.HTTPError(404)
            spider = session.query(Spider).filter_by(id=job.spider_id).first()
            storage = FilesystemEggStorage({})
            version, f = storage.get(spider.project.name)
            self.write(f.read())
            session.close()




from sqlalchemy import *
from migrate import *
meta = MetaData()

def upgrade(migrate_engine):
    meta.bind = migrate_engine
    nodes_table = Table('nodes', meta, autoload=True)
    nodes_table_is_deleted = Column('is_deleted', Integer, default=False)
    nodes_table_is_deleted.create(nodes_table)


def downgrade(migrate_engine):
    meta.bind = migrate_engine
    nodes_table = Table('nodes', meta, autoload=True)
    nodes_table.c['is_deleted'].drop()



from tornado import gen
from ..eggstorage import FilesystemEggStorage
from ..exceptions import InvalidProjectEgg, ProcessFailed
from ..workspace import ProjectWorkspace
from .base import RestBaseHandler
from ..models import session_scope, Spider, SpiderExecutionQueue, Project, SpiderSettings, NodeKey, Node
import logging
import json
from tornado.web import authenticated
from ..security import generate_digest
import hmac
from ..schedule import JobRunning
from six import BytesIO
from six import ensure_str

logger = logging.getLogger(__name__)


class GetNextJobHandler(RestBaseHandler):
    scheduler_manager = None

    def initialize(self, scheduler_manager=None):
        super(GetNextJobHandler, self).initialize()
        self.scheduler_manager = scheduler_manager

    @authenticated
    def post(self):
        with session_scope() as session:
            node_id = int(self.request.arguments['node_id'][0])
            next_task = self.scheduler_manager.get_next_task(node_id)

            response_data = {'data': None}

            if next_task is not None:
                spider = session.query(Spider).filter_by(id=next_task.spider_id).first()
                if not spider:
                    logger.error('Task %s has not spider, deleting.' % next_task.id)
                    session.query(SpiderExecutionQueue).filter_by(id=next_task.id).delete()
                    self.set_status(400)
                    self.write({'data': None})
                    return

                project = session.query(Project).filter_by(id=spider.project_id).first()
                if not project:
                    logger.error('Task %s has not project, deleting.' % next_task.id)
                    session.query(SpiderExecutionQueue).filter_by(id=next_task.id).delete()
                    self.set_status(400)
                    self.write({'data': None})
                    return

                extra_requirements_setting = session.query(SpiderSettings) \
                    .filter_by(spider_id=spider.id, setting_key='extra_requirements').first()

                extra_requirements = extra_requirements_setting.value if extra_requirements_setting else ''
                response_data['data'] = {'task': {
                    'task_id': next_task.id,
                    'spider_id': next_task.spider_id,
                    'spider_name': next_task.spider_name,
                    'project_name': next_task.project_name,
                    'version': project.version,
                    'extra_requirements': extra_requirements,
                    'spider_parameters': {parameter.parameter_key: parameter.value for parameter in spider.parameters}
                }}
            self.write(json.dumps(response_data))


class RestRegisterNodeHandler(RestBaseHandler):
    @authenticated
    def post(self):
        node_key = self.current_user
        with session_scope() as session:
            tags = self.get_argument('tags', '').strip()
            tags = None if tags == '' else tags
            remote_ip = self.request.remote_ip
            node_manager = self.settings.get('node_manager')
            node = node_manager.create_node(remote_ip, tags=tags, key_id=node_key.id)
            node_key.used_node_id = node.id
            session.add(node_key)
            session.commit()
            self.send_json({'id': node.id})


    def get_current_user(self):
        authorization = self.request.headers.get("Authorization", "").split(" ")
        if len(authorization) != 3:
            logging.info("Invalid Authorization header {}".format(authorization))
            return None

        algorithm, key, provided_digest = authorization
        if algorithm != "HMAC":
            logging.info("Invalid algorithm {}".format(algorithm))
            return None

        with session_scope() as session:
            user_key = session.query(NodeKey).filter_by(key=key).first()

            if user_key is None:
                logging.info("Invalid HMAC key {}".format(key))
                return None

            if user_key.is_deleted:
                logging.info("Invalid HMAC key {}".format(key))
                return None

            secret = user_key.secret_key
            expected_digest = generate_digest(
                secret, self.request.method, self.request.path, self.request.query,
                self.request.body)

            if not hmac.compare_digest(expected_digest, provided_digest):
                logging.info("Invalid HMAC digest {}".format(provided_digest))
                return None

            return user_key


class AddVersionHandler(RestBaseHandler):
    @authenticated
    @gen.coroutine
    def post(self):
        project_name = ensure_str(self.get_body_argument('project'))
        version = ensure_str(self.get_body_argument('version'))
        eggfile = self.request.files['egg'][0]
        eggf = BytesIO(eggfile['body'])

        try:
            workspace = ProjectWorkspace(project_name)

            yield workspace.init()
            workspace.put_egg(eggf, version)
            yield workspace.install_requirements()
            spiders = yield workspace.spider_list()

        except InvalidProjectEgg as e:
            logger.error('Error when uploading project, %s %s' % (e.message, e.detail))
            self.set_status(400, reason=e.message)
            self.finish("<html><title>%(code)d: %(message)s</title>"
                        "<body><pre>%(output)s</pre></body></html>" % {
                            "code": 400,
                            "message": e.message,
                            "output": e.detail,
                        })
            return
        except ProcessFailed as e:
            logger.error('Error when uploading project, %s, %s' % (e.message, e.std_output))
            self.set_status(400, reason=e.message)
            self.finish("<html><title>%(code)d: %(message)s</title>"
                        "<body><pre>%(output)s</pre></body></html>" % {
                            "code": 400,
                            "message": e.message,
                            "output": e.std_output,
                        })
            return

        logger.debug('spiders: %s' % spiders)
        with session_scope() as session:
            storage = FilesystemEggStorage({})
            eggf.seek(0)
            storage.put(eggf, project_name, version)
            project = session.query(Project).filter_by(name=project_name).first()
            if project is None:
                project = Project()
                project.name = project_name
            project.version = version
            session.add(project)
            session.commit()
            session.refresh(project)

            for spider_name in spiders:
                spider = session.query(Spider).filter_by(project_id=project.id, name=spider_name).first()
                if spider is None:
                    spider = Spider()
                    spider.name = spider_name
                    spider.project_id = project.id
                    session.add(spider)
                    session.commit()
                    session.refresh(spider)

                session.commit()
            self.write(json.dumps({'status': 'ok', 'spiders': len(spiders)}))


class ScheduleHandler(RestBaseHandler):
    def initialize(self, scheduler_manager):
        super(ScheduleHandler, self).initialize()
        self.scheduler_manager = scheduler_manager

    @authenticated
    def post(self):
        project = self.get_body_argument('project')
        spider = self.get_body_argument('spider')

        try:
            job = self.scheduler_manager.add_task(project, spider)
            jobid = job.id
            response_data = {
                'status': 'ok',
                'jobid': jobid
            }
            self.write(json.dumps(response_data))
        except JobRunning as e:
            response_data = {
                'status': 'error',
                'errormsg': 'job is running with jobid %s' % e.jobid
            }
            self.set_status(400, 'job is running')
            self.write(json.dumps(response_data))


from optparse import OptionParser, OptionValueError
from scrapy.utils.conf import closest_scrapy_cfg
from six.moves.configparser import ConfigParser, ParsingError
from six.moves.urllib.parse import urlencode, urljoin
from six.moves.urllib.request import urlopen
import urllib
import json

def add_schedule():
    scrapy_cfg = closest_scrapy_cfg()
    cp = ConfigParser()
    if scrapy_cfg:
        cp.read(scrapy_cfg)

    parser = OptionParser(prog  = 'scrapydd add_schedule')
    parser.add_option('-p', '--project', help='the project name')
    parser.add_option('-s', '--spider', help='the spider name')
    parser.add_option('-d', '--schedule', help='cron expression of schedule')
    parser.add_option('--host', help='the server address')
    opts, args = parser.parse_args()

    try:
        project = opts.project or cp.get('deploy', 'project')
    except ParsingError:
        print('Error: project is required')
        parser.print_help()
        return

    spider = opts.spider
    if spider is None:
        print('Error: spider is required')
        parser.print_help()
        return

    try:
        host = opts.host or cp.get('deploy', 'url')
    except Exception:
        print('Error: host is required')
        parser.print_help()
        return

    schedule = opts.schedule
    if schedule is None:
        print('Error: schedule is required')
        parser.print_help()
        return
    url = urljoin(host, '/add_schedule.json')
    postdata = urlencode({
        'project':project,
        'spider':spider,
        'cron':schedule
    })
    res = urlopen(url, postdata)
    print(res.read())



from .base import AppBaseHandler
from ..models import *
from tornado.web import authenticated
from datetime import datetime, timedelta
from ..security import generate_random_string
import uuid
from sqlalchemy import or_


class AdminNodesHandler(AppBaseHandler):
    @authenticated
    def get(self):
        if not self.current_user or not self.current_user.is_admin:
            return self.set_status(403, reason="No permission")
        with session_scope() as session:
            nodes = list(session.query(Node).filter(or_(Node.node_key_id.isnot(None),Node.is_deleted.is_(False))))
            usable_key = self._get_new_node_key()
            self.render('admin/nodes.html', nodes=nodes, usable_key=usable_key)

    def _get_new_node_key(self):
        with session_scope() as session:
            new_node_key_expire = datetime.now() - timedelta(hours=1)
            unused_node_keys = session.query(NodeKey).filter(NodeKey.used_node_id.is_(None),
                                                             NodeKey.is_deleted == False).all()
            usable_key = None
            for unused_key in unused_node_keys:
                if unused_key.create_at < new_node_key_expire:
                    unused_key.is_deleted = True
                    session.add(unused_key)
                else:
                    usable_key = unused_key
                    break

            if usable_key is None:
                usable_key = NodeKey()
                usable_key.key = str(uuid.uuid4())
                usable_key.create_at = datetime.now()
                usable_key.secret_key = generate_random_string(32)
                session.add(usable_key)
            session.commit()
            return usable_key


class AdminHomeHandler(AppBaseHandler):
    @authenticated
    def get(self):
        if not self.current_user or not self.current_user.is_admin:
            return self.set_status(403, "No permission")

        with session_scope() as session:
            user_count = session.query(User).count()
            active_node_count = session.query(Node).filter_by(isalive=True).count()
            all_node_count = session.query(Node).count()
            self.render('admin/home.html',
                        user_count = user_count,
                        active_node_count=active_node_count,
                        all_node_count=all_node_count)


from .base import AppBaseHandler, InputValidationError
from ..models import UserKey, session_scope, User
from tornado.web import authenticated
import datetime
import random
import string
import logging
from ..security import encrypt_password

logger = logging.getLogger(__name__)


def generate_random_string(length):
    letters = string.ascii_letters
    return ''.join(random.choice(letters) for _ in range(length))


class ProfileHomeHandler(AppBaseHandler):
    @authenticated
    def get(self):
        self.render('profile/profile.html')


class ProfileKeysHandler(AppBaseHandler):
    @authenticated
    def get(self):
        with session_scope() as session:
            user_keys = list(session.query(UserKey).filter_by(user_id=self.current_user.id))
            self.render('profile/profile_keys.html', user_keys=user_keys)

    @authenticated
    def post(self, *args, **kwargs):
        with session_scope() as session:
            if session.query(UserKey).filter_by(user_id=self.current_user.id).count() >= 10:
                return self.set_status(400, reason='user can only host at most 10 keys.')
            new_appkey = generate_random_string(12)
            while session.query(UserKey).filter_by(app_key=new_appkey).first():
                new_appkey = generate_random_string(12)
            new_appsecret = generate_random_string(20)

            new_key = UserKey(user_id=self.current_user.id, app_key=new_appkey, app_secret=new_appsecret,
                              create_at=datetime.datetime.now())
            session.add(new_key)
            session.commit()

            self.render('profile/profile_newkey.html', new_key=new_key)


class ProfileChangepasswordHandler(AppBaseHandler):
    @authenticated
    def get(self):
        return self.render('profile/change_password.html')

    @authenticated
    def post(self):
        try:
            old_password = self.get_body_argument('oldpassword')

            new_password = self.get_body_argument('newpassword')
            new_password2 = self.get_body_argument('newpassword2')
            if not old_password:
                raise InputValidationError("Please input password")

            encrypted_password = encrypt_password(old_password, self.settings.get('cookie_secret', ''))
            if encrypted_password != self.current_user.password:
                raise InputValidationError("Invalid password.")

            if not new_password:
                raise InputValidationError("Please input a new password.")

            if not new_password2:
                raise InputValidationError("Please confirm the new password.")

            new_password = new_password.strip()

            if new_password != new_password2:
                raise InputValidationError("Please confirm the new password.")

            encrypted_new_password = encrypt_password(new_password, self.settings.get('cookie_secret', ''))
            with session_scope() as session:
                user = session.query(User).filter_by(id=self.current_user.id).first()
                user.password =encrypted_new_password
                session.add(user)
                session.commit()

            self.render('profile/change_password_success.html')

        except InputValidationError as e:
            self.render('profile/change_password.html', error_message=e.message)


import sys
import scrapydd.main
import scrapydd.scheduleutil
import scrapydd.agent
import scrapydd.ssl_gen
from scrapydd.commands.resetpassword import ResetPasswordCommand
from scrapydd.commands.package import PackageCommand

usage = '''
usage: scrapydd {command} [options]
Available commands:
    server:         run scrapydd server.
    agent:          run scrapydd agent.
    cert:           make certs.
    add_schedule:   add a schedule to spider.
    reset_password: reset user password.
    package:        package current project egg.

use scrapydd {command} --help for further help.
'''

def main():
    argv = sys.argv
    if len(argv) == 1:
        print('No command specified.')
        print_commands()
        sys.exit(1)
        return
    cmd = argv[1]
    if cmd == 'agent':
        scrapydd.agent.run(argv)
    elif cmd == 'server':
        scrapydd.main.run(argv)
    elif cmd == 'add_schedule':
        scrapydd.scheduleutil.add_schedule()
    elif cmd == 'cert':
        scrapydd.ssl_gen.run(argv[2:])
    elif cmd == 'reset_password':
        ResetPasswordCommand().run()
    elif cmd == 'package':
        PackageCommand().run()
    elif cmd == '--help':
        print_commands(argv)
    elif cmd == '--version' or cmd == '-v':
        print_version()
    else:
        print('Invalid command.')
        print_commands()

def print_commands():
    print(usage)

def print_version():
    print('scrapydd version:', scrapydd.__version__)

if __name__ == '__main__':
    main()


from tornado.process import cpu_count, _reseed_random
from tornado import ioloop
import logging
import os
import signal
import sys
from tornado.util import errno_from_exception
import errno

logger = logging.getLogger(__name__)

_task_id = None
exiting = False

def fork_processes(num_processes, max_restarts=100):
    """Starts multiple worker processes.

    If ``num_processes`` is None or <= 0, we detect the number of cores
    available on this machine and fork that number of child
    processes. If ``num_processes`` is given and > 0, we fork that
    specific number of sub-processes.

    Since we use processes and not threads, there is no shared memory
    between any server code.

    Note that multiple processes are not compatible with the autoreload
    module (or the ``autoreload=True`` option to `tornado.web.Application`
    which defaults to True when ``debug=True``).
    When using multiple processes, no IOLoops can be created or
    referenced until after the call to ``fork_processes``.

    In each child process, ``fork_processes`` returns its *task id*, a
    number between 0 and ``num_processes``.  Processes that exit
    abnormally (due to a signal or non-zero exit status) are restarted
    with the same id (up to ``max_restarts`` times).  In the parent
    process, ``fork_processes`` returns None if all child processes
    have exited normally, but will otherwise only exit by throwing an
    exception.
    """
    global _task_id
    assert _task_id is None
    if num_processes is None or num_processes <= 0:
        num_processes = cpu_count()
    if ioloop.IOLoop.initialized():
        raise RuntimeError("Cannot run in multiple processes: IOLoop instance "
                           "has already been initialized. You cannot call "
                           "IOLoop.instance() before calling start_processes()")
    logger.info("Starting %d processes", num_processes)
    children = {}

    def start_child(i):
        pid = os.fork()
        if pid == 0:
            # child process
            _reseed_random()
            global _task_id
            _task_id = i
            return i
        else:
            children[pid] = i
            return None

    for i in range(num_processes):
        id = start_child(i)
        if id is not None:
            return id
    global exiting
    exiting = False
    def receive_signal(sig, frame):
        logger.debug('Received signal')
        global exiting
        exiting = True
        for pid, taskid in children.items():
            os.kill(pid, signal.SIGTERM)


    signal.signal(signal.SIGTERM, receive_signal)
    signal.signal(signal.SIGINT, receive_signal)
    num_restarts = 0
    while children and not exiting:
        logger.debug('Exiting : %s' % exiting)
        try:
            pid, status = os.wait()
        except OSError as e:
            if errno_from_exception(e) == errno.EINTR:
                continue
            raise
        if pid not in children:
            continue
        id = children.pop(pid)
        if os.WIFSIGNALED(status):
            logger.warning("child %d (pid %d) killed by signal %d, restarting",
                            id, pid, os.WTERMSIG(status))
        elif os.WEXITSTATUS(status) != 0:
            logger.warning("child %d (pid %d) exited with status %d, restarting",
                            id, pid, os.WEXITSTATUS(status))
        else:
            logger.info("child %d (pid %d) exited normally", id, pid)
            continue
        num_restarts += 1
        if num_restarts > max_restarts:
            raise RuntimeError("Too many child restarts, giving up")
        new_id = start_child(id)
        if new_id is not None:
            return new_id
    # All child processes exited cleanly, so exit the master process
    # instead of just returning to right after the call to
    # fork_processes (which will probably just start up another IOLoop
    # unless the caller checks the return value).
    sys.exit(0)


from sqlalchemy import *
from migrate import *
meta = MetaData()

node_key_table = Table(
    'node_keys', meta,
    Column('id', Integer, primary_key=True),
    Column('key', String(length=50), unique=True, nullable=False),
    Column('secret_key', String(length=50), nullable=False),
    Column('is_deleted', Boolean, default=False),
    Column('used_node_id', Integer),
    Column('create_at', DateTime, nullable=False),
)


def upgrade(migrate_engine):
    meta.bind = migrate_engine
    node_key_table.create()

    nodes_table = Table('nodes', meta, autoload=True)
    nodes_table_key_id = Column('node_key_id', Integer, ForeignKey('node_keys.id'))
    #nodes_table_is_deleted = Column('is_deleted', Boolean, default=False)
    nodes_table_key_id.create(nodes_table)
    #nodes_table_is_deleted.create(nodes_table)


def downgrade(migrate_engine):
    meta.bind = migrate_engine
    node_key_table.drop()

    nodes_table = Table('nodes', meta, autoload=True)
    nodes_table.c['node_key_id'].drop()
    #nodes_table.c['is_deleted'].drop()


import functools
from .models import User, UserKey, session_scope
from six.moves.urllib.parse import urlsplit, urlencode
from six.moves.urllib_parse import urlparse
from tornado.web import HTTPError
from tornado.httpclient import HTTPRequest
import logging
import hmac
import hashlib
import string
import random
import base64
from six import ensure_binary, ensure_str

logger = logging.getLogger(__name__)

try:
    from urllib.parse import parse_qs, quote
except ImportError:
    from urlparse import parse_qs
    from urllib import quote

try:
    from base64 import decodebytes as decodebytes
except ImportError:
    from base64 import decodestring as decodebytes


def generate_random_string(length):
    letters = string.ascii_letters
    return ''.join(random.choice(letters) for _ in range(length))


def generate_digest(secret, method, path, query, body):
    parsed_query = parse_qs(query, keep_blank_values=True)

    canonical_query = []

    for key in sorted(parsed_query.keys()):
        for value in sorted(parsed_query[key]):
            canonical_query.append("=".join((key, quote(value))))
    body = ensure_binary(body)
    return hmac.new(
        secret.encode("utf-8"),
        "\n".join((method, path, "&".join(canonical_query), "")).encode("utf-8") +
        body,
        hashlib.sha256).hexdigest()


def signin_view(func):
    pass


def authentication_check(method):
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self.current_user:
            if self.request.method in ("GET", "HEAD"):
                url = self.get_login_url()
                if "?" not in url:
                    if urlsplit(url).scheme:
                        # if login url is absolute, make next absolute too
                        next_url = self.request.full_url()
                    else:
                        assert self.request.uri is not None
                        next_url = self.request.uri
                    url += "?" + urlencode(dict(next=next_url))
                self.redirect(url)
                return None
            raise HTTPError(403)
        return method(self, *args, **kwargs)
    return wrapper


class AuthenticationProvider(object):
    def get_user(self, handler):
        raise NotImplementedError()


class NoAuthenticationProvider(object):
    def get_user(self, handler):
        return 'admin'


class CookieAuthenticationProvider(object):
    def get_user(self, handler):
        user_cookie = handler.get_secure_cookie("user")
        if user_cookie:
            #return json.loads(user_cookie)
            return ensure_str(user_cookie)
        return None


class HmacAuthorize(object):
    def get_user(self, handler):
        authorization = handler.request.headers.get("Authorization", "").split(" ")
        if len(authorization) != 3:
            logging.info("Invalid Authorization header {}".format(authorization))
            return None

        algorithm, key, provided_digest = authorization
        if algorithm != "HMAC":
            logging.info("Invalid algorithm {}".format(algorithm))
            return None

        with session_scope() as session:
            user_key = session.query(UserKey).filter_by(app_key=key).first()

            if user_key is None:
                logging.info("Invalid HMAC key {}".format(key))
                return None
            secret = user_key.app_secret
            expected_digest = generate_digest(
                secret, handler.request.method, handler.request.path, handler.request.query,
                handler.request.body)

            if not hmac.compare_digest(expected_digest, provided_digest):
                logging.info("Invalid HMAC digest {}".format(provided_digest))
                return None

            return user_key.user.username


class BasicAuthentication(object):
    def get_user(self, handler):
        logger.debug('BasicAuthentication.get_user')
        authorization_header = handler.request.headers.get("Authorization")
        if authorization_header is None:
            return
        authorization = ensure_str(authorization_header).split(" ")
        if len(authorization) != 2:
            logging.info("Invalid Authorization header {}".format(authorization))
            return None

        algorithm, encrypted = authorization
        if algorithm != 'Basic':
            return None

        decrypted = ensure_str(decodebytes(ensure_binary(encrypted)))
        username, password = decrypted.split(':', 1)
        with session_scope() as session:
            user = session.query(User).filter_by(username=username).first()
            if not user:
                return None

            encrypted_password = encrypt_password(password, handler.settings.get('cookie_secret', ''))
            if user.password == encrypted_password:
                return username

        return None


def authenticated_request(*args, **kwargs):
    app_key = kwargs.pop("app_key")
    app_secret = kwargs.pop("app_secret")

    if len(args) > 0:
        url = args[0]
    elif "url" in kwargs:
        url = kwargs["url"]
    else:
        raise TypeError("Missing argument: 'url'")

    parsed_url = urlparse(url)

    path = parsed_url.path
    query = parsed_url.query

    body = kwargs.get("body", "")
    if isinstance(body, str):
        body = body.encode("utf-8")

    digest = generate_digest(app_secret, kwargs.get("method", "GET"), path, query, body)

    headers = kwargs.get("headers", {})
    headers["Authorization"] = "HMAC {} {}".format(app_key, digest)
    kwargs["headers"] = headers

    return HTTPRequest(*args, **kwargs)

def encrypt_password(origin_password, salt):
    hash = hashlib.sha1()
    hash.update(ensure_binary(salt))
    hash.update(ensure_binary(origin_password))
    return hash.hexdigest()



from .base import AppBaseHandler
from ..models import session_scope, User
from ..security import encrypt_password


class SignupHandler(AppBaseHandler):
    def get(self):
        pass

    def post(self):
        pass


class SigninHandler(AppBaseHandler):
    def get(self):
        self.render('signin.html')

    def post(self):
        username = self.get_body_argument('username')
        password = self.get_body_argument('password')

        with session_scope() as session:
            user = session.query(User).filter_by(username=username).first()
            if user is None:
                return self.get()

            encrypted_password = encrypt_password(password, self.settings.get('cookie_secret', ''))

            if user.password != encrypted_password:
                return self.get()

            self.set_secure_cookie("user", username)
            self.redirect(self.get_query_argument('next', '/'))


class LogoutHandler(AppBaseHandler):
    def get(self):
        self.clear_cookie('user')
        self.redirect('/')



#!/usr/bin/env python
import sys, os

'''This module is used to fork the current process into a daemon.
    Almost none of this is necessary (or advisable) if your daemon
    is being started by inetd. In that case, stdin, stdout and stderr are
    all set up for you to refer to the network connection, and the fork()s
    and session manipulation should not be done (to avoid confusing inetd).
    Only the chdir() and umask() steps remain as useful.
    References:
        UNIX Programming FAQ
            1.7 How do I get my program to act like a daemon?
                http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16

        Advanced Programming in the Unix Environment
            W. Richard Stevens, 1992, Addison-Wesley, ISBN 0-201-56317-7.
    '''

def daemonize (stdin='/dev/null', stdout='/dev/null', stderr='/dev/null', chdir=None, pidfile=None):
    '''This forks the current process into a daemon.
    The stdin, stdout, and stderr arguments are file names that
    will be opened and be used to replace the standard file descriptors
    in sys.stdin, sys.stdout, and sys.stderr.
    These arguments are optional and default to /dev/null.
    Note that stderr is opened unbuffered, so
    if it shares a file with stdout then interleaved output
    may not appear in the order that you expect.
    '''
    # Do first fork.
    try:
        pid = os.fork()
        if pid > 0:
            sys.exit(0) # Exit first parent.
    except OSError as e:
        sys.stderr.write ('fork #1 failed: (%d) %s\n' % (e.errno, e.strerror)    )
        sys.exit(1)

    # Decouple from parent environment.
    if chdir is not None:
        os.chdir(chdir)
    os.umask(0)
    os.setsid()

    # Do second fork.
    try:
        pid = os.fork()
        if pid > 0:
            sys.exit(0) # Exit second parent.
    except OSError as e:
        sys.stderr.write ('fork #2 failed: (%d) %s\n' % (e.errno, e.strerror)    )
        sys.exit(1)
    if pidfile:
        with open(pidfile, 'w+') as f_pidfile:
            f_pidfile.write(str(os.getpid()))

    # Now I am a daemon!

    # Redirect standard file descriptors.
    si = file(stdin, 'r')
    so = file(stdout, 'a+')
    se = file(stderr, 'a+', 0)
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())


from setuptools import setup, find_packages
import os
import scrapydd
import io

version = scrapydd.__version__

def read_file(filename):
    with io.open(filename) as fp:
        return fp.read().strip()

def read_requirements(filename):
    return [line.strip() for line in read_file(filename).splitlines()
            if not line.startswith('#')]

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))
setup(
    name         = 'scrapydd',
    version      = version,
    author       ='kevenli',
    author_email ='pbleester@gmail.com',
    url          = 'http://github.com/kevenli/scrapydd',
    packages     = find_packages(exclude=('tests', 'tests.*')),
    description  = 'distributed scrapy spider scheduling system',
    long_description=read_file('README.rst'),
    long_description_content_type="text/x-rst",
    package_data = {
        'scrapydd': [
            'scrapydd.default.conf',
            'scrapyddagent.default.conf',
            'templates/**/*.html',
            'templates/*.html',
            'migrates/migrate.cfg',
            'static/**/*',
        ],
    },
    entry_points = {
        'console_scripts': [
            'scrapydd = scrapydd.scripts.cmdline:main',
        ]
    },
    install_requires=read_requirements('requirements.txt'),
)
